@@ -1,5 +1,5 @@
 /*
-   Copyright 2005-2010 Jakub Kruszona-Zawadzki, Gemius SA, 2013 Skytechnology sp. z o.o..
+   Copyright 2005-2010 Jakub Kruszona-Zawadzki, Gemius SA, 2013-2014 EditShare, 2013-2015 Skytechnology sp. z o.o..
 
    This file was part of MooseFS and is part of LizardFS.
 
@@ -21,6 +21,9 @@
 #include "common/platform.h"
 
 #include <inttypes.h>
+#include <array>
+#include <string>
+#include <vector>
 
 typedef struct _md5ctx {
 	uint32_t state[4];
@@ -31,3 +34,7 @@ typedef struct _md5ctx {
 void md5_init(md5ctx *ctx);
 void md5_update(md5ctx *ctx,const uint8_t *buff,uint32_t leng);
 void md5_final(uint8_t digest[16],md5ctx *ctx);
+std::array<uint8_t, 16> md5_challenge_response(const std::array<uint8_t, 32>& challenge,
+		std::string data);
+
+int md5_parse(std::vector<uint8_t> &password_digest, const char *in_md5_data);
