@@ -1,5 +1,5 @@
 /*
-   Copyright 2005-2010 Jakub Kruszona-Zawadzki, Gemius SA, 2013 Skytechnology sp. z o.o..
+   Copyright 2005-2010 Jakub Kruszona-Zawadzki, Gemius SA, 2013-2014 EditShare, 2013-2015 Skytechnology sp. z o.o..
 
    This file was part of MooseFS and is part of LizardFS.
 
@@ -22,8 +22,17 @@
 
 #include <inttypes.h>
 
+inline void queue_deleter_dummy(uint8_t *) {
+}
+
+template<typename T>
+void queue_deleter_delete(uint8_t *p) {
+	delete ((T*)p);
+}
+
+
 void* queue_new(uint32_t size);
-void queue_delete(void *que);
+void queue_delete(void *que, void(*deleter)(uint8_t *) = queue_deleter_dummy);
 int queue_isempty(void *que);
 uint32_t queue_elements(void *que);
 int queue_isfull(void *que);
