@@ -1,3 +1,21 @@
+/*
+   Copyright 2013-2014 EditShare, 2013-2015 Skytechnology sp. z o.o.
+
+   This file is part of LizardFS.
+
+   LizardFS is free software: you can redistribute it and/or modify
+   it under the terms of the GNU General Public License as published by
+   the Free Software Foundation, version 3.
+
+   LizardFS is distributed in the hope that it will be useful,
+   but WITHOUT ANY WARRANTY; without even the implied warranty of
+   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
+   GNU General Public License for more details.
+
+   You should have received a copy of the GNU General Public License
+   along with LizardFS. If not, see <http://www.gnu.org/licenses/>.
+ */
+
 #pragma once
 
 #include "common/platform.h"
@@ -11,10 +29,13 @@
 #include <utility>
 #include <vector>
 
-
+#include "common/compact_vector.h"
+#include "common/flat_set.h"
+#include "common/flat_map.h"
 #include "common/datapack.h"
 #include "common/exception.h"
 #include "common/massert.h"
+#include "common/small_vector.h"
 
 const uint32_t kMaxDeserializedBytesCount = 32 * 1024 * 1024;  // 32MiB
 const uint32_t kMaxDeserializedElementsCount = 1000 * 1000;    // 1M
@@ -30,46 +51,46 @@ public:
 
 // serializedSize
 
-inline uint32_t serializedSize(const bool&) {
+inline uint32_t serializedSize(bool) {
 	return 1;
 }
 
-inline uint32_t serializedSize(const uint8_t&) {
+inline uint32_t serializedSize(char) {
 	return 1;
 }
 
-inline uint32_t serializedSize(const uint16_t&) {
+inline uint32_t serializedSize(signed char) {
+	return 1;
+}
+
+inline uint32_t serializedSize(unsigned char) {
+	return 1;
+}
+
+inline uint32_t serializedSize(uint16_t) {
 	return 2;
 }
 
-inline uint32_t serializedSize(const uint32_t&) {
+inline uint32_t serializedSize(uint32_t) {
 	return 4;
 }
 
-inline uint32_t serializedSize(const uint64_t&) {
+inline uint32_t serializedSize(uint64_t) {
 	return 8;
 }
 
-inline uint32_t serializedSize(const int8_t&) {
-	return 1;
-}
-
-inline uint32_t serializedSize(const int16_t&) {
+inline uint32_t serializedSize(int16_t) {
 	return 2;
 }
 
-inline uint32_t serializedSize(const int32_t&) {
+inline uint32_t serializedSize(int32_t) {
 	return 4;
 }
 
-inline uint32_t serializedSize(const int64_t&) {
+inline uint32_t serializedSize(int64_t) {
 	return 8;
 }
 
-inline uint32_t serializedSize(const char& c) {
-	return serializedSize(reinterpret_cast<const uint8_t&>(c));
-}
-
 inline uint32_t serializedSize(const std::string& value) {
 	return serializedSize(uint32_t(value.size())) + value.size() + 1;
 }
@@ -82,12 +103,16 @@ template<class T>
 inline uint32_t serializedSize(const std::unique_ptr<T>& ptr);
 template<class T, class A>
 inline uint32_t serializedSize(const std::vector<T, A>& vector);
+template<class T, size_t Size>
+inline uint32_t serializedSize(const small_vector<T, Size> &vector);
 template <typename K, typename C, typename A>
 inline uint32_t serializedSize(const std::set<K, C, A>& set);
 template<class T1, class T2>
 inline uint32_t serializedSize(const std::pair<T1, T2>& pair);
 template <typename K, typename T, typename C, typename A>
 inline uint32_t serializedSize(const std::map<K, T, C, A>& map);
+template<class T>
+inline uint32_t serializedSize(const T& t);
 template<class T, class ... Args>
 inline uint32_t serializedSize(const T& t, const Args& ... args);
 
@@ -133,6 +158,16 @@ inline uint32_t serializedSize(const std::vector<T, A>& vector) {
 	return ret;
 }
 
+template<class T, size_t Size>
+inline uint32_t serializedSize(const small_vector<T, Size> &vector) {
+	uint32_t ret = 0;
+	ret += serializedSize(uint32_t(vector.size()));
+	for (const auto& t : vector) {
+		ret += serializedSize(t);
+	}
+	return ret;
+}
+
 template <typename K, typename C, typename A>
 inline uint32_t serializedSize(const std::set<K, C, A>& set) {
 	uint32_t ret = 0;
@@ -153,6 +188,26 @@ inline uint32_t serializedSize(const std::map<K, T, C, A>& map) {
 	return ret;
 }
 
+template <typename T, typename Size, typename Alloc>
+inline uint32_t serializedSize(const compact_vector<T, Size, Alloc>& vector) {
+	uint32_t ret = 0;
+	ret += serializedSize(uint32_t(vector.size()));
+	for (const auto& t : vector) {
+		ret += serializedSize(t);
+	}
+	return ret;
+}
+
+template <typename T, typename C, class Compare>
+inline uint32_t serializedSize(const flat_set<T, C, Compare>& set) {
+	return serializedSize(set.data());
+}
+
+template <typename Key, typename T, typename C, class Compare>
+inline uint32_t serializedSize(const flat_map<Key, T, C, Compare>& map) {
+	return serializedSize(map.data());
+}
+
 template<class T>
 inline uint32_t serializedSize(const T& t) {
 	return t.serializedSize();
@@ -166,44 +221,44 @@ inline uint32_t serializedSize(const T& t, const Args& ... args) {
 // serialize for simple types
 
 // serialize bool
-inline void serialize(uint8_t** destination, const bool& value) {
+inline void serialize(uint8_t** destination, bool value) {
 	put8bit(destination, static_cast<uint8_t>(value ? 1 : 0));
 }
 
-inline void serialize(uint8_t** destination, const uint8_t& value) {
+inline void serialize(uint8_t** destination, char value) {
 	put8bit(destination, value);
 }
 
-inline void serialize(uint8_t** destination, const uint16_t& value) {
-	put16bit(destination, value);
-}
-
-inline void serialize(uint8_t** destination, const uint32_t& value) {
-	put32bit(destination, value);
-}
-
-inline void serialize(uint8_t** destination, const uint64_t& value) {
-	put64bit(destination, value);
-}
-
-inline void serialize(uint8_t** destination, const int8_t& value) {
+inline void serialize(uint8_t** destination, signed char value) {
 	put8bit(destination, value);
 }
 
-inline void serialize(uint8_t** destination, const int16_t& value) {
+inline void serialize(uint8_t** destination, unsigned char value) {
+	put8bit(destination, value);
+}
+
+inline void serialize(uint8_t** destination, uint16_t value) {
 	put16bit(destination, value);
 }
 
-inline void serialize(uint8_t** destination, const int32_t& value) {
+inline void serialize(uint8_t** destination, uint32_t value) {
 	put32bit(destination, value);
 }
 
-inline void serialize(uint8_t** destination, const int64_t& value) {
+inline void serialize(uint8_t** destination, uint64_t value) {
 	put64bit(destination, value);
 }
 
-inline void serialize(uint8_t** destination, const char& value) {
-	serialize(destination, reinterpret_cast<const uint8_t&>(value));
+inline void serialize(uint8_t** destination, int16_t value) {
+	put16bit(destination, value);
+}
+
+inline void serialize(uint8_t** destination, int32_t value) {
+	put32bit(destination, value);
+}
+
+inline void serialize(uint8_t** destination, int64_t value) {
+	put64bit(destination, value);
 }
 
 // serialize a string
@@ -224,10 +279,14 @@ template<class T>
 inline void serialize(uint8_t** destination, const std::unique_ptr<T>& ptr);
 template<class T, class A>
 inline void serialize(uint8_t** destination, const std::vector<T, A>& vector);
+template<class T, size_t Size>
+inline void serialize(uint8_t **destination, const small_vector<T, Size> &vector);
 template <typename K, typename C, typename A>
 inline void serialize(uint8_t** destination, const std::set<K, C, A>& set);
 template <typename K, typename T, typename C, typename A>
 inline void serialize(uint8_t** destination, const std::map<K, T, C, A>& map);
+template<class T>
+inline void serialize(uint8_t** destination, const T& t);
 template<class T, class... Args>
 inline void serialize(uint8_t** destination, const T& t, const Args&... args);
 
@@ -273,6 +332,14 @@ inline void serialize(uint8_t** destination, const std::vector<T, A>& vector) {
 	}
 }
 
+template<class T, size_t Size>
+inline void serialize(uint8_t **destination, const small_vector<T, Size> &vector) {
+	serialize(destination, uint32_t(vector.size()));
+	for (const T& t : vector) {
+		serialize(destination, t);
+	}
+}
+
 template <typename K, typename C, typename A>
 inline void serialize(uint8_t** destination, const std::set<K, C, A>& set) {
 	serialize(destination, uint32_t(set.size()));
@@ -289,6 +356,23 @@ inline void serialize(uint8_t** destination, const std::map<K, T, C, A>& map) {
 	}
 }
 
+template <typename T, typename Size, typename Alloc>
+inline void serialize(uint8_t** destination, const compact_vector<T, Size, Alloc>& vector) {
+	serialize(destination, uint32_t(vector.size()));
+	for (const T& t : vector) {
+		serialize(destination, t);
+	}
+}
+
+template <typename T, typename C, class Compare>
+inline void serialize(uint8_t** destination, const flat_set<T, C, Compare>& set) {
+	serialize(destination, set.data());
+}
+template <typename Key, typename T, typename C, class Compare>
+inline void serialize(uint8_t** destination, const flat_map<Key, T, C, Compare>& map) {
+	serialize(destination, map.data());
+}
+
 // serialization
 template<class T>
 inline void serialize(uint8_t** destination, const T& t) {
@@ -322,12 +406,26 @@ inline void deserialize(const uint8_t** source, uint32_t& bytesLeftInBuffer, boo
 	value = static_cast<bool>(integerValue);
 }
 
-inline void deserialize(const uint8_t** source, uint32_t& bytesLeftInBuffer, uint8_t& value) {
+inline void deserialize(const uint8_t** source, uint32_t& bytesLeftInBuffer, char& value) {
 	verifySize(value, bytesLeftInBuffer);
 	bytesLeftInBuffer -= 1;
 	value = get8bit(source);
 }
 
+inline void deserialize(const uint8_t** source, uint32_t& bytesLeftInBuffer, signed char& value) {
+	verifySize(value, bytesLeftInBuffer);
+	bytesLeftInBuffer -= 1;
+	value = get8bit(source);
+}
+
+
+inline void deserialize(const uint8_t** source, uint32_t& bytesLeftInBuffer, unsigned char& value) {
+	verifySize(value, bytesLeftInBuffer);
+	bytesLeftInBuffer -= 1;
+	value = get8bit(source);
+}
+
+
 inline void deserialize(const uint8_t** source, uint32_t& bytesLeftInBuffer, uint16_t& value) {
 	verifySize(value, bytesLeftInBuffer);
 	bytesLeftInBuffer -= 2;
@@ -346,12 +444,6 @@ inline void deserialize(const uint8_t** source, uint32_t& bytesLeftInBuffer, uin
 	value = get64bit(source);
 }
 
-inline void deserialize(const uint8_t** source, uint32_t& bytesLeftInBuffer, int8_t& value) {
-	verifySize(value, bytesLeftInBuffer);
-	bytesLeftInBuffer -= 1;
-	value = get8bit(source);
-}
-
 inline void deserialize(const uint8_t** source, uint32_t& bytesLeftInBuffer, int16_t& value) {
 	verifySize(value, bytesLeftInBuffer);
 	bytesLeftInBuffer -= 2;
@@ -370,10 +462,6 @@ inline void deserialize(const uint8_t** source, uint32_t& bytesLeftInBuffer, int
 	value = get64bit(source);
 }
 
-inline void deserialize(const uint8_t** source, uint32_t& bytesLeftInBuffer, char& value) {
-	deserialize(source, bytesLeftInBuffer, reinterpret_cast<uint8_t&>(value));
-}
-
 // deserialize a string
 inline void deserialize(const uint8_t** source, uint32_t& bytesLeftInBuffer, std::string& value) {
 	sassert(value.size() == 0);
@@ -410,12 +498,17 @@ inline void deserialize(const uint8_t** source, uint32_t& bytesLeftInBuffer,
 template<class T, class A>
 inline void deserialize(const uint8_t** source, uint32_t& bytesLeftInBuffer,
 		std::vector<T, A>& vec);
+template<class T, size_t Size>
+inline void deserialize(const uint8_t **source, uint32_t& bytesLeftInBuffer,
+		small_vector<T, Size> &vec);
 template <typename K, typename C, typename A>
 inline void deserialize(const uint8_t** source, uint32_t& bytesLeftInBuffer,
 		std::set<K, C, A>& set);
 template <typename K, typename T, typename C, typename A>
 inline void deserialize(const uint8_t** source, uint32_t& bytesLeftInBuffer,
 		std::map<K, T, C, A>& map);
+template<class T>
+inline void deserialize(const uint8_t** source, uint32_t& bytesLeftInBuffer, T& t);
 template<class T, class... Args>
 inline void deserialize(const uint8_t** source, uint32_t& bytesLeftInBuffer, T& t, Args&... args);
 
@@ -480,6 +573,21 @@ inline void deserialize(const uint8_t** source, uint32_t& bytesLeftInBuffer,
 	}
 }
 
+template<class T, size_t Size>
+inline void deserialize(const uint8_t **source, uint32_t& bytesLeftInBuffer,
+		small_vector<T, Size> &vec) {
+	sassert(vec.size() == 0);
+	uint32_t size;
+	deserialize(source, bytesLeftInBuffer, size);
+	if (size > kMaxDeserializedElementsCount) {
+		throw IncorrectDeserializationException("untrustworthy vector size");
+	}
+	vec.resize(size);
+	for (unsigned i = 0; i < size; ++i) {
+		deserialize(source, bytesLeftInBuffer, vec[i]);
+	}
+}
+
 template <typename K, typename C, typename A>
 inline void deserialize(const uint8_t** source, uint32_t& bytesLeftInBuffer,
 		std::set<K, C, A>& set) {
@@ -512,6 +620,33 @@ inline void deserialize(const uint8_t** source, uint32_t& bytesLeftInBuffer,
 	}
 }
 
+template<typename T, typename Size, typename Alloc>
+inline void deserialize(const uint8_t** source, uint32_t& bytesLeftInBuffer,
+		compact_vector<T, Size, Alloc>& vec) {
+	sassert(vec.size() == 0);
+	uint32_t size;
+	deserialize(source, bytesLeftInBuffer, size);
+	if (size > kMaxDeserializedElementsCount) {
+		throw IncorrectDeserializationException("untrustworthy compact_vector size");
+	}
+	vec.resize(size);
+	for(auto &element : vec) {
+		deserialize(source, bytesLeftInBuffer, element);
+	}
+}
+
+template <typename T, typename C, class Compare>
+inline void deserialize(const uint8_t** source, uint32_t& bytesLeftInBuffer,
+		flat_set<T, C, Compare>& set) {
+	deserialize(source, bytesLeftInBuffer, set.data());
+}
+
+template <typename Key, typename T, typename C, class Compare>
+inline void deserialize(const uint8_t** source, uint32_t& bytesLeftInBuffer,
+		flat_map<Key, T, C, Compare>& map) {
+	deserialize(source, bytesLeftInBuffer, map.data());
+}
+
 template<class T>
 inline void deserialize(const uint8_t** source, uint32_t& bytesLeftInBuffer, T& t) {
 	return t.deserialize(source, bytesLeftInBuffer);
