@@ -1,5 +1,5 @@
 /*
-   Copyright 2005-2010 Jakub Kruszona-Zawadzki, Gemius SA, 2013 Skytechnology sp. z o.o..
+   Copyright 2005-2010 Jakub Kruszona-Zawadzki, Gemius SA, 2013-2014 EditShare, 2013-2015 Skytechnology sp. z o.o..
 
    This file was part of MooseFS and is part of LizardFS.
 
@@ -19,27 +19,27 @@
 #include "common/platform.h"
 #include "init.h"
 
-#ifndef MFSMAXFILES
-#define MFSMAXFILES 5000
-#endif
-
 #include <errno.h>
 #include <fcntl.h>
 #include <grp.h>
+#include <ios>
 #include <pwd.h>
 #include <signal.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 #include <strings.h>
-#include <syslog.h>
 #include <sys/resource.h>
 #include <sys/stat.h>
 #include <sys/time.h>
 #include <sys/types.h>
 #include <sys/wait.h>
+#include <syslog.h>
 #include <time.h>
 #include <unistd.h>
+#include <algorithm>
+#include <atomic>
+#include <fstream>
 #include <list>
 #include <memory>
 
@@ -48,11 +48,14 @@
 #include "common/cwrap.h"
 #include "common/exceptions.h"
 #include "common/exit_status.h"
+#include "common/event_loop.h"
 #include "common/main.h"
 #include "common/massert.h"
 #include "common/mfserr.h"
 #include "common/setup.h"
 #include "common/slogger.h"
+#include "common/time_utils.h"
+#include "protocol/MFSCommunication.h"
 
 #if defined(LIZARDFS_HAVE_MLOCKALL)
 #  if defined(LIZARDFS_HAVE_SYS_MMAN_H)
@@ -66,6 +69,19 @@
 #  endif
 #endif
 
+#if defined(LIZARDFS_HAVE_PAM)
+#  include <security/pam_appl.h>
+#  include <security/pam_misc.h>
+#endif
+
+#ifndef LIZARDFS_MAX_FILES
+#  define LIZARDFS_MAX_FILES 5000
+#endif
+
+#if defined(LIZARDFS_HAVE_SYSTEMD_SD_DAEMON_H)
+#include <systemd/sd-daemon.h>
+#endif
+
 #define STR_AUX(x) #x
 #define STR(x) STR_AUX(x)
 
@@ -79,408 +95,72 @@ enum class RunMode {
 	kIsAlive = 6
 };
 
-static bool nextPollNonblocking = false;
+#if defined(LIZARDFS_HAVE_PAM)
+static pam_handle_t *gPAMHandle=NULL;
+#endif
 
-typedef struct deentry {
-	void (*fun)(void);
-	struct deentry *next;
-} deentry;
-
-static deentry *dehead=NULL;
-
-
-typedef struct weentry {
-	void (*fun)(void);
-	struct weentry *next;
-} weentry;
-
-static weentry *wehead=NULL;
-
-
-typedef struct ceentry {
-	int (*fun)(void);
-	struct ceentry *next;
-} ceentry;
-
-static ceentry *cehead=NULL;
-
-
-typedef struct rlentry {
-	void (*fun)(void);
-	struct rlentry *next;
-} rlentry;
-
-static rlentry *rlhead=NULL;
-
-
-typedef struct pollentry {
-	void (*desc)(struct pollfd *,uint32_t *);
-	void (*serve)(struct pollfd *);
-	struct pollentry *next;
-} pollentry;
-
-static pollentry *pollhead=NULL;
-
-
-typedef struct eloopentry {
-	void (*fun)(void);
-	struct eloopentry *next;
-} eloopentry;
-
-static eloopentry *eloophead=NULL;
-
-
-struct timeentry {
-	typedef void (*fun_t)(void);
-	timeentry(uint32_t ne, uint32_t sec, uint32_t off, int mod, fun_t f)
-		: nextevent(ne), seconds(sec), offset(off), mode(mod), fun(f) {
-	}
-	uint32_t nextevent;
-	uint32_t seconds;
-	uint32_t offset;
-	int mode;
-	fun_t fun;
-};
-
-typedef std::list<timeentry> TimeEntries;
-namespace {
-TimeEntries gTimeEntries;
-}
-
-static uint32_t now;
-static uint64_t usecnow;
 static bool gRunAsDaemon = true;
 
+/// When set to true, config will be reloaded after the current loop
+
 static int signalpipe[2];
 
 /* interface */
 
-void main_make_next_poll_nonblocking() {
-	nextPollNonblocking = true;
+static void signal_pipe_desc(std::vector<pollfd> &pdesc) {
+	assert(pdesc.size() == 0);
+	pdesc.push_back({signalpipe[0],POLLIN,0});
 }
 
-void main_destructregister (void (*fun)(void)) {
-	deentry *aux=(deentry*)malloc(sizeof(deentry));
-	passert(aux);
-	aux->fun = fun;
-	aux->next = dehead;
-	dehead = aux;
-}
-
-void main_canexitregister (int (*fun)(void)) {
-	ceentry *aux=(ceentry*)malloc(sizeof(ceentry));
-	passert(aux);
-	aux->fun = fun;
-	aux->next = cehead;
-	cehead = aux;
-}
-
-void main_wantexitregister (void (*fun)(void)) {
-	weentry *aux=(weentry*)malloc(sizeof(weentry));
-	passert(aux);
-	aux->fun = fun;
-	aux->next = wehead;
-	wehead = aux;
-}
-
-void main_reloadregister (void (*fun)(void)) {
-	rlentry *aux=(rlentry*)malloc(sizeof(rlentry));
-	passert(aux);
-	aux->fun = fun;
-	aux->next = rlhead;
-	rlhead = aux;
-}
-
-void main_pollregister (void (*desc)(struct pollfd *,uint32_t *),void (*serve)(struct pollfd *)) {
-	pollentry *aux=(pollentry*)malloc(sizeof(pollentry));
-	passert(aux);
-	aux->desc = desc;
-	aux->serve = serve;
-	aux->next = pollhead;
-	pollhead = aux;
-}
-
-void main_eachloopregister (void (*fun)(void)) {
-	eloopentry *aux=(eloopentry*)malloc(sizeof(eloopentry));
-	passert(aux);
-	aux->fun = fun;
-	aux->next = eloophead;
-	eloophead = aux;
-}
-
-void* main_timeregister (int mode,uint32_t seconds,uint32_t offset,void (*fun)(void)) {
-	if (seconds==0 || offset>=seconds) {
-		return NULL;
-	}
-	uint32_t nextevent = ((now / seconds) * seconds) + offset;
-	while (nextevent<now) {
-		nextevent+=seconds;
-	}
-	gTimeEntries.push_front(timeentry(nextevent, seconds, offset, mode, fun));
-	return &gTimeEntries.front();
-}
-
-void main_timeunregister(void* handler) {
-	TimeEntries::iterator it(gTimeEntries.begin());
-	for (TimeEntries::iterator end(gTimeEntries.end()); it != end; ++it) {
-		if (&(*it) == handler) {
-			gTimeEntries.erase(it);
-			break;
-		}
-	}
-	massert(it != gTimeEntries.end(), "unregistering unknown handle from time table");
-}
-
-int main_timechange(void* x,int mode,uint32_t seconds,uint32_t offset) {
-	timeentry *aux = (timeentry*)x;
-	if (seconds==0 || offset>=seconds) {
-		return -1;
-	}
-	aux->nextevent = ((now / seconds) * seconds) + offset;
-	while (aux->nextevent<now) {
-		aux->nextevent+=seconds;
-	}
-	aux->seconds = seconds;
-	aux->offset = offset;
-	aux->mode = mode;
-	return 0;
-}
-
-/* internal */
-
-void free_all_registered_entries(void) {
-	deentry *de,*den;
-	ceentry *ce,*cen;
-	weentry *we,*wen;
-	rlentry *re,*ren;
-	pollentry *pe,*pen;
-	eloopentry *ee,*een;
-
-	for (de = dehead ; de ; de = den) {
-		den = de->next;
-		free(de);
-	}
-
-	for (ce = cehead ; ce ; ce = cen) {
-		cen = ce->next;
-		free(ce);
-	}
-
-	for (we = wehead ; we ; we = wen) {
-		wen = we->next;
-		free(we);
-	}
-
-	for (re = rlhead ; re ; re = ren) {
-		ren = re->next;
-		free(re);
-	}
-
-	for (pe = pollhead ; pe ; pe = pen) {
-		pen = pe->next;
-		free(pe);
-	}
-
-	for (ee = eloophead ; ee ; ee = een) {
-		een = ee->next;
-		free(ee);
-	}
-
-	gTimeEntries.clear();
-}
-
-int canexit() {
-	ceentry *aux;
-	for (aux = cehead ; aux!=NULL ; aux=aux->next) {
-		if (aux->fun()==0) {
-			return 0;
-		}
-	}
-	return 1;
-}
-
-uint32_t main_time() {
-	return now;
-}
-
-uint64_t main_utime() {
-	return usecnow;
-}
-
-void destruct() {
-	deentry *deit;
-	for (deit = dehead ; deit!=NULL ; deit=deit->next) {
-		deit->fun();
-	}
-}
-
-void mainloop() {
-	uint32_t prevtime = 0;
-	struct timeval tv;
-	pollentry *pollit;
-	eloopentry *eloopit;
-	ceentry *ceit;
-	weentry *weit;
-	rlentry *rlit;
-	struct pollfd pdesc[MFSMAXFILES];
-	uint32_t ndesc;
-	int i;
-	int t,r;
-
-	t = 0;
-	r = 0;
-	while (t!=3) {
-		ndesc=1;
-		pdesc[0].fd = signalpipe[0];
-		pdesc[0].events = POLLIN;
-		pdesc[0].revents = 0;
-		for (pollit = pollhead ; pollit != NULL ; pollit = pollit->next) {
-			pollit->desc(pdesc,&ndesc);
-		}
-		i = poll(pdesc,ndesc, nextPollNonblocking ? 0 : 50);
-		nextPollNonblocking = false;
-		gettimeofday(&tv,NULL);
-		usecnow = tv.tv_sec;
-		usecnow *= 1000000;
-		usecnow += tv.tv_usec;
-		now = tv.tv_sec;
-		if (i<0) {
-			if (errno==EAGAIN) {
-				syslog(LOG_WARNING,"poll returned EAGAIN");
-				usleep(100000);
-				continue;
-			}
-			if (errno!=EINTR) {
-				syslog(LOG_WARNING,"poll error: %s",strerr(errno));
-				break;
-			}
-		} else {
-			if ((pdesc[0].revents)&POLLIN) {
-				uint8_t sigid;
-				if (read(signalpipe[0],&sigid,1)==1) {
-					if (sigid=='\001' && t==0) {
-						syslog(LOG_NOTICE,"terminate signal received");
-						t = 1;
-					} else if (sigid=='\002') {
-						syslog(LOG_NOTICE,"reloading config files");
-						r = 1;
-					} else if (sigid=='\003') {
-						syslog(LOG_NOTICE, "Received SIGUSR1, killing gently...");
-						exit(LIZARDFS_EXIT_STATUS_GENTLY_KILL);
-					}
-				}
-			}
-			for (pollit = pollhead ; pollit != NULL ; pollit = pollit->next) {
-				pollit->serve(pdesc);
-			}
-		}
-		for (eloopit = eloophead ; eloopit != NULL ; eloopit = eloopit->next) {
-			eloopit->fun();
-		}
-		if (now<prevtime) {
-			// time went backward !!! - recalculate "nextevent" time
-			// adding previous_time_to_run prevents from running next event too soon.
-			for (timeentry& timeit : gTimeEntries) {
-				uint32_t previous_time_to_run = timeit.nextevent - prevtime;
-				if (previous_time_to_run > timeit.seconds) {
-					previous_time_to_run = timeit.seconds;
-				}
-				timeit.nextevent = ((now / timeit.seconds) * timeit.seconds) + timeit.offset;
-				while (timeit.nextevent <= now+previous_time_to_run) {
-					timeit.nextevent += timeit.seconds;
-				}
-			}
-		} else if (now>prevtime+3600) {
-			// time went forward !!! - just recalculate "nextevent" time
-			for (timeentry& timeit : gTimeEntries) {
-				timeit.nextevent = ((now / timeit.seconds) * timeit.seconds) + timeit.offset;
-				while (now >= timeit.nextevent) {
-					timeit.nextevent += timeit.seconds;
-				}
-			}
-		}
-		for (timeentry& timeit : gTimeEntries) {
-			if (now >= timeit.nextevent) {
-				if (timeit.mode == TIMEMODE_RUN_LATE) {
-					while (now >= timeit.nextevent) {
-						timeit.nextevent += timeit.seconds;
-					}
-					timeit.fun();
-				} else { /* timeit.mode == TIMEMODE_SKIP_LATE */
-					if (now == timeit.nextevent) {
-						timeit.fun();
-					}
-					while (now >= timeit.nextevent) {
-						timeit.nextevent += timeit.seconds;
-					}
-				}
-			}
-		}
-		prevtime = now;
-		if (t==0 && r) {
-			cfg_reload();
-			for (rlit = rlhead ; rlit!=NULL ; rlit=rlit->next) {
-				try {
-					rlit->fun();
-				} catch (Exception& ex) {
-					syslog(LOG_WARNING, "reload error: %s", ex.what());
-				}
-			}
-			r = 0;
-			DEBUG_LOG("main.reload");
-		}
-		if (t==1) {
-			for (weit = wehead ; weit!=NULL ; weit=weit->next) {
-				weit->fun();
-			}
-			t = 2;
-		}
-		if (t==2) {
-			i = 1;
-			for (ceit = cehead ; ceit!=NULL && i ; ceit=ceit->next) {
-				if (ceit->fun()==0) {
-					i=0;
-				}
-			}
-			if (i) {
-				t = 3;
+static void signal_pipe_serv(const std::vector<pollfd> &pdesc) {
+	if ((pdesc[0].revents)&POLLIN) {
+		uint8_t sigid;
+		if (read(signalpipe[0],&sigid,1)==1) {
+			if (sigid == '\001' && gExitingStatus == ExitingStatus::kRunning) {
+				lzfs_pretty_syslog(LOG_NOTICE,"terminate signal received");
+				gExitingStatus = ExitingStatus::kWantExit;
+			} else if (sigid=='\002') {
+				lzfs_pretty_syslog(LOG_NOTICE,"reloading config files");
+				gReloadRequested = true;
+			} else if (sigid=='\003') {
+				lzfs_pretty_syslog(LOG_NOTICE, "Received SIGUSR1, killing gently...");
+				exit(LIZARDFS_EXIT_STATUS_GENTLY_KILL);
 			}
 		}
 	}
 }
 
-int initialize(void) {
+
+int initialize(run_tab* tab) {
 	uint32_t i;
 	int ok;
 	ok = 1;
-	for (i=0 ; (long int)(RunTab[i].fn)!=0 && ok ; i++) {
-		now = time(NULL);
+	for (i=0 ; (long int)(tab[i].fn)!=0 && ok ; i++) {
+		eventloop_updatetime();
 		try {
-			if (RunTab[i].fn() < 0) {
-				mfs_arg_syslog(LOG_ERR, "init: %s failed !!!", RunTab[i].name);
+			if (tab[i].fn()<0) {
+				lzfs_pretty_syslog(LOG_ERR,"init: %s failed",tab[i].name);
 				ok=0;
 			}
 		} catch (const std::exception& e) {
-			mfs_arg_syslog(LOG_ERR, "init: %s failed: %s !!!", RunTab[i].name, e.what());
+			lzfs_pretty_syslog(LOG_ERR, "%s", e.what());
 			ok = 0;
 		}
 	}
+	eventloop_updatetime();
 	return ok;
 }
 
+int initialize_early(void) {
+	return initialize(EarlyRunTab);
+}
+
+int initialize(void) {
+	return initialize(RunTab);
+}
+
 int initialize_late(void) {
-	uint32_t i;
-	int ok;
-	ok = 1;
-	for (i=0 ; (long int)(LateRunTab[i].fn)!=0 && ok ; i++) {
-		now = time(NULL);
-		if (LateRunTab[i].fn()<0) {
-			mfs_arg_syslog(LOG_ERR,"init: %s failed !!!",RunTab[i].name);
-			ok=0;
-		}
-	}
-	now = time(NULL);
-	return ok;
+	return initialize(LateRunTab);
 }
 
 const std::string& set_syslog_ident() {
@@ -499,33 +179,44 @@ const std::string& set_syslog_ident() {
 		openlog(logIdent.c_str(), LOG_PID | LOG_NDELAY, LOG_USER);
 #endif
 	}
+
 	return logIdent;
 }
 
 static void main_configure_debug_log() {
-	std::string debugLogConfiguration;
+	std::string flush_on_str = cfg_getstring("LOG_FLUSH_ON", "CRITICAL");
+	int priority = LOG_CRIT;
+	if (flush_on_str == "ERROR") {
+		priority = LOG_ERR;
+	} else if (flush_on_str == "WARNING") {
+		priority = LOG_WARNING;
+	} else if (flush_on_str == "INFO") {
+		priority = LOG_INFO;
+	} else if (flush_on_str == "DEBUG") {
+		priority = LOG_DEBUG;
+	}
+	lzfs::drop_all_logs();
+	lzfs::add_log_syslog();
 	for (std::string suffix : {"", "_A", "_B", "_C"}) {
 		std::string configEntryName = "MAGIC_DEBUG_LOG" + suffix;
 		std::string value = cfg_get(configEntryName.c_str(), "");
 		if (value.empty()) {
 			continue;
 		}
-		if (!debugLogConfiguration.empty()) {
-			debugLogConfiguration += ',';
-		}
-		debugLogConfiguration += value;
+		lzfs_add_log_file(value.c_str(), LOG_DEBUG, 16*1024*1024, 8);
 	}
-	DebugLog::setConfiguration(std::move(debugLogConfiguration));
+	lzfs_set_log_flush_on(priority);
 }
 
 void main_reload() {
 	// Reload SYSLOG_IDENT
-	syslog(LOG_NOTICE, "Changing SYSLOG_IDENT to %s",
+	lzfs_pretty_syslog(LOG_NOTICE, "Changing SYSLOG_IDENT to %s",
 			cfg_get("SYSLOG_IDENT", STR(APPNAME)).c_str());
 	set_syslog_ident();
 
 	// Reload MAGIC_DEBUG_LOG
 	main_configure_debug_log();
+	lzfs_silent_syslog(LOG_DEBUG, "main.reload");
 }
 
 /* signals */
@@ -639,7 +330,7 @@ void signal_cleanup(void) {
 	close(signalpipe[1]);
 }
 
-void changeugid(void) {
+void changeugid(RunMode runmode) {
 	char pwdgrpbuff[16384];
 	struct passwd pwd,*pw;
 	struct group grp,*gr;
@@ -661,7 +352,7 @@ void changeugid(void) {
 		} else if (wgroup[0]) {
 			getgrnam_r(wgroup,&grp,pwdgrpbuff,16384,&gr);
 			if (gr==NULL) {
-				mfs_arg_syslog(LOG_WARNING,"%s: no such group !!!",wgroup);
+				lzfs_pretty_syslog(LOG_WARNING,"%s: no such group",wgroup);
 				exit(LIZARDFS_EXIT_STATUS_ERROR);
 			} else {
 				wrk_gid = gr->gr_gid;
@@ -674,7 +365,7 @@ void changeugid(void) {
 			if (gidok==0) {
 				getpwuid_r(wrk_uid,&pwd,pwdgrpbuff,16384,&pw);
 				if (pw==NULL) {
-					mfs_arg_syslog(LOG_ERR,"%s: no such user id - can't obtain group id",wuser+1);
+					lzfs_pretty_syslog(LOG_ERR,"%s: no such user id - can't obtain group id",wuser+1);
 					exit(LIZARDFS_EXIT_STATUS_ERROR);
 				}
 				wrk_gid = pw->pw_gid;
@@ -682,7 +373,7 @@ void changeugid(void) {
 		} else {
 			getpwnam_r(wuser,&pwd,pwdgrpbuff,16384,&pw);
 			if (pw==NULL) {
-				mfs_arg_syslog(LOG_ERR,"%s: no such user !!!",wuser);
+				lzfs_pretty_syslog(LOG_ERR,"%s: no such user",wuser);
 				exit(LIZARDFS_EXIT_STATUS_ERROR);
 			}
 			wrk_uid = pw->pw_uid;
@@ -693,21 +384,61 @@ void changeugid(void) {
 		free(wuser);
 		free(wgroup);
 
-		if (setgid(wrk_gid)<0) {
-			mfs_arg_errlog(LOG_ERR,"can't set gid to %d",(int)wrk_gid);
+		if (setgroups(0, NULL) < 0) {
+			lzfs::log_warn("can't reset supplementary groups");
+		}
+		if (setgid(wrk_gid) < 0) {
+			lzfs::log_err("can't set gid to {}", (int)wrk_gid);
 			exit(LIZARDFS_EXIT_STATUS_ERROR);
-		} else {
-			syslog(LOG_NOTICE,"set gid to %d",(int)wrk_gid);
+		} else if ((runmode == RunMode::kStart) || (runmode == RunMode::kRestart)){
+			lzfs::log_info("set gid to {}", (int)wrk_gid);
 		}
 		if (setuid(wrk_uid)<0) {
-			mfs_arg_errlog(LOG_ERR,"can't set uid to %d",(int)wrk_uid);
+			lzfs::log_err("can't set uid to {}", (int)wrk_uid);
 			exit(LIZARDFS_EXIT_STATUS_ERROR);
-		} else {
-			syslog(LOG_NOTICE,"set uid to %d",(int)wrk_uid);
+		} else if ((runmode == RunMode::kStart) || (runmode == RunMode::kRestart)){
+			lzfs::log_info("set uid to {}", (int)wrk_uid);
 		}
 	}
 }
 
+bool open_pam_session() {
+#if defined(LIZARDFS_HAVE_PAM)
+	static struct pam_conv conv = {misc_conv, NULL};
+
+	assert(gPAMHandle == NULL);
+
+	int retval;
+
+	retval = pam_start("lizardfs", DEFAULT_USER, &conv, &gPAMHandle);
+	if (retval != PAM_SUCCESS) {
+		lzfs_pretty_errlog(LOG_ERR, "Can't initialize PAM");
+		return false;
+	}
+
+	retval = pam_open_session(gPAMHandle, 0);
+	if (retval != PAM_SUCCESS) {
+		lzfs_pretty_errlog(LOG_ERR, "Can't open PAM session");
+		return false;
+	}
+
+	return true;
+#else
+	return false;
+#endif
+}
+
+void close_pam_session() {
+#if defined(LIZARDFS_HAVE_PAM)
+	if (!gPAMHandle) {
+		return;
+	}
+
+	int retval = pam_close_session(gPAMHandle, 0);
+	pam_end(gPAMHandle, retval);
+#endif
+}
+
 class FileLock {
 public:
 	enum class LockStatus {
@@ -724,18 +455,15 @@ public:
 			  thisProcessCreatedLockFile_(false) {
 		while ((lockstatus_ = wdlock(runmode, timeout)) == LockStatus::kAgain) {
 		}
-		if (lockstatus_ == LockStatus::kFail) {
-			throw FilesystemException("");
-		}
 	}
-	virtual ~FileLock() {
+	~FileLock() {
 		/*
 		 * Order of unlock/file deletion matters
 		 * for RunMode::kRestart mode.
 		 * We need to remove file first and then
 		 * unlock file.
 		 */
-		if (thisProcessCreatedLockFile_) {
+		if (lockstatus_ != LockStatus::kFail && thisProcessCreatedLockFile_) {
 			::unlink(name_.c_str());
 		}
 	}
@@ -750,7 +478,7 @@ public:
 private:
 	LockStatus wdlock(RunMode runmode, uint32_t timeout);
 	pid_t mylock();
-	void createLockFile();
+	bool createLockFile();
 
 	FileDescriptor fd_;
 	std::string name_;
@@ -785,22 +513,32 @@ pid_t FileLock::mylock() {
 	return pid;
 }
 
-void FileLock::createLockFile() {
+bool FileLock::createLockFile() {
 	bool notExisted((::access(name_.c_str(), F_OK) != 0) && (errno == ENOENT));
 	fd_.reset(open(name_.c_str(), O_WRONLY | O_CREAT, 0666));
 	if (!fd_.isOpened()) {
-		throw FilesystemException("can't create lockfile in working directory");
+		std::string err = "can't create lockfile " + fs::getCurrentWorkingDirectoryNoThrow()
+		                  + "/" + name();
+		lzfs_pretty_errlog(LOG_ERR, "%s", err.c_str());
+
+		return false;
 	}
 	thisProcessCreatedLockFile_ = notExisted;
+	return true;
 }
 
 FileLock::LockStatus FileLock::wdlock(RunMode runmode, uint32_t timeout) {
-	createLockFile();
-	pid_t ownerpid(mylock());
-	if (ownerpid<0) {
-		mfs_errlog(LOG_ERR,"fcntl error");
+	std::string lockPath = fs::getCurrentWorkingDirectoryNoThrow() + "/" + name_;
+	if(!createLockFile()) {
 		return LockStatus::kFail;
 	}
+
+	pid_t ownerpid(mylock());
+	if (ownerpid<0) {
+		lzfs_pretty_errlog(LOG_ERR, "fcntl error while creating lockfile %s", lockPath.c_str());
+		return LockStatus::kFail;
+	}
+
 	if (ownerpid>0) {
 		if (runmode==RunMode::kIsAlive) {
 			return LockStatus::kAlive;
@@ -810,7 +548,9 @@ FileLock::LockStatus FileLock::wdlock(RunMode runmode, uint32_t timeout) {
 			return LockStatus::kFail;
 		}
 		if (runmode==RunMode::kStart) {
-			fprintf(stderr,"can't start: lockfile is already locked by another process\n");
+			lzfs_pretty_syslog(LOG_ERR,
+					"can't start: lockfile %s is already locked by another process",
+					lockPath.c_str());
 			return LockStatus::kFail;
 		}
 		if (runmode==RunMode::kReload) {
@@ -818,14 +558,15 @@ FileLock::LockStatus FileLock::wdlock(RunMode runmode, uint32_t timeout) {
 			 * FIXME: buissiness logic should not be in file locking function.
 			 */
 			if (kill(ownerpid,SIGHUP)<0) {
-				mfs_errlog(LOG_WARNING,"can't send reload signal to lock owner");
+				lzfs_pretty_errlog(LOG_ERR,
+						"can't send reload signal to the lock owner of %s",
+						lockPath.c_str());
 				return LockStatus::kFail;
 			}
-			fprintf(stderr,"reload signal has beed sent\n");
+			lzfs_pretty_syslog(LOG_INFO, "reload signal has been sent");
 			return LockStatus::kSuccess;
 		}
 		if (runmode==RunMode::kKill) {
-			fprintf(stderr,"sending SIGKILL to lock owner (pid:%ld)\n",(long int)ownerpid);
 			/*
 			 * FIXME: buissiness logic should not be in file locking function.
 			 */
@@ -834,21 +575,23 @@ FileLock::LockStatus FileLock::wdlock(RunMode runmode, uint32_t timeout) {
 #else
 			if (kill(ownerpid,SIGKILL)<0) {
 #endif
-				mfs_errlog(LOG_WARNING,"can't kill lock owner");
+				lzfs_pretty_errlog(LOG_ERR,
+						"can't kill (SIGKILL) the lock owner (pid:%ld) of %s",
+						(long int)ownerpid, lockPath.c_str());
 				return LockStatus::kFail;
 			}
 		} else {
 			sassert((runmode == RunMode::kStop) || (runmode == RunMode::kRestart));
-			fprintf(stderr,"sending SIGTERM to lock owner (pid:%ld)\n",(long int)ownerpid);
 			/*
 			 * FIXME: buissiness logic should not be in file locking function.
 			 */
 			if (kill(ownerpid,SIGTERM)<0) {
-				mfs_errlog(LOG_WARNING,"can't kill lock owner");
+				lzfs_pretty_errlog(LOG_ERR,
+						"can't kill (SIGTERM) the lock owner (pid:%ld) of %s",
+						(long int)ownerpid, lockPath.c_str());
 				return LockStatus::kFail;
 			}
 		}
-		fprintf(stderr,"waiting for termination ... ");
 		fflush(stderr);
 		uint32_t l = 0;
 		pid_t newownerpid;
@@ -856,40 +599,41 @@ FileLock::LockStatus FileLock::wdlock(RunMode runmode, uint32_t timeout) {
 		do {
 			newownerpid = mylock();
 			if (newownerpid<0) {
-				mfs_errlog(LOG_ERR,"fcntl error");
+				lzfs_pretty_errlog(LOG_ERR,
+						"fcntl error while creating lockfile %s",
+						lockPath.c_str());
 				return LockStatus::kFail;
 			}
 			if (newownerpid>0) {
 				l++;
 				uint32_t secondsElapsed = l / checksPerSecond;
 				if (secondsElapsed >= timeout) {
-					syslog(LOG_ERR,
-							"about %" PRIu32 " seconds passed and lock still exists - giving up",
-							secondsElapsed);
-					fprintf(stderr,"giving up\n");
+					lzfs_pretty_syslog(LOG_ERR,
+							"about %" PRIu32 " seconds passed and the lock %s still exists - giving up",
+							secondsElapsed, lockPath.c_str());
 					return LockStatus::kFail;
 				}
 				if (l % (10 * checksPerSecond) == 0) {
-					syslog(LOG_WARNING, "about %" PRIu32 " seconds passed and lock still exists",
-							secondsElapsed);
-					fprintf(stderr,"%" PRIu32 "s ", secondsElapsed);
+					lzfs_pretty_syslog(LOG_WARNING,
+							"about %" PRIu32 " seconds passed and the lock %s still exists...",
+							secondsElapsed, lockPath.c_str());
 					fflush(stderr);
 				}
 				if (newownerpid!=ownerpid) {
-					fprintf(stderr,"\nnew lock owner detected\n");
+					lzfs_silent_syslog(LOG_INFO, "new lock owner of %s detected", lockPath.c_str());
 					if (runmode==RunMode::kKill) {
-						fprintf(stderr,"sending SIGKILL to lock owner (pid:%ld) ... ",(long int)newownerpid);
-						fflush(stderr);
 						if (kill(newownerpid,SIGKILL)<0) {
-							mfs_errlog(LOG_WARNING,"can't kill lock owner");
+							lzfs_pretty_errlog(LOG_ERR,
+									"can't kill (SIGKILL) the lock owner (pid:%ld) of %s",
+									(long int)newownerpid, lockPath.c_str());
 							return LockStatus::kFail;
 						}
 					} else {
 						sassert((runmode == RunMode::kStop) || (runmode == RunMode::kRestart));
-						fprintf(stderr,"sending SIGTERM to lock owner (pid:%ld) ... ",(long int)newownerpid);
-						fflush(stderr);
 						if (kill(newownerpid,SIGTERM)<0) {
-							mfs_errlog(LOG_WARNING,"can't kill lock owner");
+							lzfs_pretty_errlog(LOG_ERR,
+									"can't kill (SIGTERM) lock owner (pid:%ld) of %s",
+									(long int)newownerpid, lockPath.c_str());
 							return LockStatus::kFail;
 						}
 					}
@@ -898,19 +642,18 @@ FileLock::LockStatus FileLock::wdlock(RunMode runmode, uint32_t timeout) {
 			}
 			usleep(1000000 / checksPerSecond);
 		} while (newownerpid!=0);
-		fprintf(stderr,"terminated\n");
 		return (runmode == RunMode::kRestart) ? LockStatus::kAgain : LockStatus::kSuccess;
 	}
 	if (runmode==RunMode::kStart || runmode==RunMode::kRestart) {
-		fprintf(stderr,"lockfile created and locked\n");
+		lzfs_pretty_syslog(LOG_INFO,"lockfile %s created and locked", lockPath.c_str());
 	} else if (runmode==RunMode::kStop || runmode==RunMode::kKill) {
-		fprintf(stderr,"can't find process to terminate\n");
+		lzfs_pretty_syslog(LOG_WARNING,"can't find process to terminate");
 		return LockStatus::kSuccess;
 	} else if (runmode==RunMode::kReload) {
-		fprintf(stderr,"can't find process to send reload signal\n");
+		lzfs_pretty_syslog(LOG_ERR,"can't find process to send reload signal");
 		return LockStatus::kFail;
 	} else if (runmode==RunMode::kTest) {
-		fprintf(stderr,STR(APPNAME) " is not running\n");
+		fprintf(stderr, STR(APPNAME) " is not running\n");
 	} else if (runmode==RunMode::kIsAlive) {
 		return LockStatus::kSuccess;
 	}
@@ -927,18 +670,18 @@ void makedaemon() {
 	fflush(stdout);
 	fflush(stderr);
 	if (pipe(piped)<0) {
-		fprintf(stderr,"pipe error\n");
+		lzfs_pretty_syslog(LOG_ERR, "pipe error");
 		exit(LIZARDFS_EXIT_STATUS_ERROR);
 	}
 	f = fork();
 	if (f<0) {
-		syslog(LOG_ERR,"first fork error: %s",strerr(errno));
+		lzfs_pretty_errlog(LOG_ERR, "first fork error");
 		exit(LIZARDFS_EXIT_STATUS_ERROR);
 	}
 	if (f>0) {
 		wait(&f);       // just get child status - prevents child from being zombie during initialization stage
 		if (f) {
-			fprintf(stderr,"Child status: %d\n",f);
+			lzfs_pretty_syslog(LOG_ERR, "child status: %d",f);
 			exit(LIZARDFS_EXIT_STATUS_ERROR);
 		}
 		close(piped[1]);
@@ -954,7 +697,7 @@ void makedaemon() {
 				happy = fwrite(pipebuff,1,r,stderr);
 				(void)happy;
 			} else {
-				fprintf(stderr,"Error reading pipe: %s\n",strerr(errno));
+				lzfs_pretty_errlog(LOG_ERR,"error reading pipe");
 				exit(LIZARDFS_EXIT_STATUS_ERROR);
 			}
 		}
@@ -964,9 +707,9 @@ void makedaemon() {
 	setpgid(0,getpid());
 	f = fork();
 	if (f<0) {
-		syslog(LOG_ERR,"second fork error: %s",strerr(errno));
+		lzfs_pretty_errlog(LOG_ERR,"second fork error");
 		if (write(piped[1],"fork error\n",11)!=11) {
-			syslog(LOG_ERR,"pipe write error: %s",strerr(errno));
+			lzfs_pretty_errlog(LOG_ERR,"pipe write error");
 		}
 		close(piped[1]);
 		exit(LIZARDFS_EXIT_STATUS_ERROR);
@@ -983,6 +726,15 @@ void makedaemon() {
 	close(STDERR_FILENO);
 	sassert(dup(piped[1])==STDERR_FILENO);
 	close(piped[1]);
+
+	// close all inherited file descriptors
+	int open_max = sysconf(_SC_OPEN_MAX);
+	for (int i = 3; i < open_max; i++) {
+		if (i == signalpipe[0] || i == signalpipe[1]) {
+			continue;
+		}
+		close(i);
+	}
 }
 
 void close_msg_channel() {
@@ -1005,10 +757,10 @@ void createpath(const char *filename) {
 			*dst='\0';
 			if (mkdir(pathbuff,(mode_t)0777)<0) {
 				if (errno!=EEXIST) {
-					mfs_arg_errlog(LOG_NOTICE,"creating directory %s",pathbuff);
+					lzfs_pretty_errlog(LOG_NOTICE,"creating directory %s",pathbuff);
 				}
 			} else {
-				mfs_arg_syslog(LOG_NOTICE,"directory %s has been created",pathbuff);
+				lzfs_pretty_syslog(LOG_NOTICE,"directory %s has been created",pathbuff);
 			}
 			*dst++=*src++;
 		}
@@ -1024,48 +776,49 @@ void usage(const char *appname) {
 "-u : log undefined config variables\n"
 "-t locktimeout : how long wait for lockfile\n"
 "-c cfgfile : use given config file\n"
+"-p pidfile : write pid to given file\n"
+"-o extra_option : module specific extra option\n"
 	,appname);
 	exit(LIZARDFS_EXIT_STATUS_ERROR);
 }
 
+void makePidFile(const std::string &name) {
+	if (name.empty()) {
+		return;
+	}
+
+	std::ofstream ofs(name, std::ios_base::out | std::ios_base::trunc);
+	if (ofs.fail()) {
+		lzfs_pretty_syslog(LOG_WARNING, "failed to create pid file: %s",
+				name.c_str());
+		return;
+	}
+	ofs << std::to_string(getpid()) << std::endl;
+}
+
 int main(int argc,char **argv) {
 	char *wrkdir;
-	char *cfgfile;
 	char *appname;
 	int ch;
 	int logundefined;
 	int lockmemory;
 	int32_t nicelevel;
 	uint32_t locktimeout;
-	int fd;
-	uint8_t movewarning;
 	struct rlimit rls;
+	std::string default_cfgfile = ETC_PATH "/" STR(APPNAME) ".cfg";
+	std::string cfgfile = default_cfgfile;
+	std::string pidfile;
 
 	prepareEnvironment();
-	strerr_init();
 	mycrc32_init();
 
-	movewarning = 0;
-	cfgfile=strdup(ETC_PATH "/mfs/" STR(APPNAME) ".cfg");
-	passert(cfgfile);
-	if ((fd = open(cfgfile,O_RDONLY))<0 && errno==ENOENT) {
-		free(cfgfile);
-		cfgfile=strdup(ETC_PATH "/" STR(APPNAME) ".cfg");
-		passert(cfgfile);
-		if ((fd = open(cfgfile,O_RDONLY))>=0) {
-			movewarning = 1;
-		}
-	}
-	if (fd>=0) {
-		close(fd);
-	}
 	locktimeout = 1800;
 	RunMode runmode = RunMode::kRestart;
 	logundefined = 0;
 	lockmemory = 0;
 	appname = argv[0];
 
-	while ((ch = getopt(argc, argv, "c:dht:uvx?")) != -1) {
+	while ((ch = getopt(argc, argv, "o:c:p:dht:uvx?")) != -1) {
 		switch(ch) {
 			case 'v':
 				printf("version: %s\n",LIZARDFS_PACKAGE_VERSION);
@@ -1076,11 +829,14 @@ int main(int argc,char **argv) {
 			case 't':
 				locktimeout=strtoul(optarg,NULL,10);
 				break;
+			case 'o':
+				gExtraArguments.emplace_back(optarg);
+				break;
 			case 'c':
-				free(cfgfile);
-				cfgfile = strdup(optarg);
-				passert(cfgfile);
-				movewarning = 0;
+				cfgfile = optarg;
+				break;
+			case 'p':
+				pidfile = optarg;
 				break;
 			case 'u':
 				logundefined=1;
@@ -1119,30 +875,44 @@ int main(int argc,char **argv) {
 		return LIZARDFS_EXIT_STATUS_ERROR;
 	}
 
-	if (movewarning) {
-		mfs_syslog(LOG_WARNING,"default sysconf path has changed - please move " STR(APPNAME) ".cfg from " ETC_PATH "/ to " ETC_PATH "/mfs/");
+	if (gRunAsDaemon) {
+		lzfs::add_log_stderr(lzfs::log_level::warn);
+	} else {
+		lzfs::add_log_stderr(lzfs::log_level::debug);
 	}
-
 	if (runmode==RunMode::kStart || runmode==RunMode::kRestart) {
 		if (gRunAsDaemon) {
 			makedaemon();
 		} else {
 			set_signal_handlers(0);
 		}
+		makePidFile(pidfile);
 	}
 
-	if (cfg_load(cfgfile,logundefined)==0) {
-		fprintf(stderr,"can't load config file: %s - using defaults\n",cfgfile);
+	if (cfg_load(cfgfile.c_str(), logundefined)==0) {
+		lzfs_pretty_syslog(LOG_WARNING, "configuration file %s not found - using defaults; "
+				"please create one to remove this warning "
+				"(you can copy %s.dist to get a base configuration)",
+				cfgfile.c_str(), default_cfgfile.c_str());
+	} else if (runmode==RunMode::kStart || runmode==RunMode::kRestart) {
+		lzfs_pretty_syslog(LOG_INFO, "configuration file %s loaded", cfgfile.c_str());
 	}
-	free(cfgfile);
 	main_configure_debug_log();
-	const std::string& logappname = set_syslog_ident();
 
 	if (runmode==RunMode::kStart || runmode==RunMode::kRestart) {
-		rls.rlim_cur = MFSMAXFILES;
-		rls.rlim_max = MFSMAXFILES;
-		if (setrlimit(RLIMIT_NOFILE,&rls)<0) {
-			syslog(LOG_NOTICE,"can't change open files limit to %u",MFSMAXFILES);
+		if (!open_pam_session()) {
+			rls.rlim_cur = LIZARDFS_MAX_FILES;
+			rls.rlim_max = LIZARDFS_MAX_FILES;
+			if (setrlimit(RLIMIT_NOFILE, &rls) < 0) {
+				lzfs_pretty_syslog(
+				    LOG_WARNING,
+				    "can't change open files limit to %u, check restrictions with `ulimit -n'",
+				    LIZARDFS_MAX_FILES);
+			} else {
+				lzfs_pretty_syslog(
+				    LOG_INFO, "failed to use system open files limit, changed to default value %u",
+				    LIZARDFS_MAX_FILES);
+			}
 		}
 
 		lockmemory = cfg_getnum("LOCK_MEMORY",0);
@@ -1157,36 +927,42 @@ int main(int argc,char **argv) {
 		setpriority(PRIO_PROCESS,getpid(),nicelevel);
 	}
 
-	changeugid();
+	changeugid(runmode);
 
 	wrkdir = cfg_getstr("DATA_PATH",DATA_PATH);
-	if (runmode==RunMode::kStart || runmode==RunMode::kRestart) {
-		fprintf(stderr,"working directory: %s\n",wrkdir);
-	}
+
 
 	if (chdir(wrkdir)<0) {
-		mfs_arg_syslog(LOG_ERR,"can't set working directory to %s",wrkdir);
+		lzfs_pretty_syslog(LOG_ERR,"can't set working directory to %s",wrkdir);
 		if (gRunAsDaemon) {
 			fputc(0,stderr);
 			close_msg_channel();
 		}
 		closelog();
 		return LIZARDFS_EXIT_STATUS_ERROR;
+	} else {
+		if (runmode==RunMode::kStart || runmode==RunMode::kRestart) {
+			lzfs_pretty_syslog(LOG_INFO,"changed working directory to: %s",wrkdir);
+		}
 	}
 	free(wrkdir);
 
 	umask(cfg_getuint32("FILE_UMASK",027)&077);
 
-	std::unique_ptr<FileLock> fl;
-	try {
-		/*
-		 * Only kStart should check for lock file consistency.
-		 */
-		fl.reset(new FileLock(runmode, locktimeout));
-	} catch (const FilesystemException& e) {
-		if (e.what()[0]) {
-			mfs_errlog(LOG_ERR, e.what());
+	eventloop_pollregister(signal_pipe_desc, signal_pipe_serv);
+
+	if (!initialize_early()) {
+		if (gRunAsDaemon) {
+			fputc(0, stderr);
+			close_msg_channel();
 		}
+		closelog();
+		return LIZARDFS_EXIT_STATUS_ERROR;
+	}
+
+	// Only kStart should check for lock file consistency
+	FileLock fl(runmode, locktimeout);
+	if (fl.lockstatus() == FileLock::LockStatus::kFail) {
 		if (gRunAsDaemon) {
 			fputc(0,stderr);
 			close_msg_channel();
@@ -1202,7 +978,7 @@ int main(int argc,char **argv) {
 		}
 		closelog();
 		if (runmode==RunMode::kIsAlive) {
-			FileLock::LockStatus lockstatus = fl->lockstatus();
+			FileLock::LockStatus lockstatus = fl.lockstatus();
 			sassert((lockstatus == FileLock::LockStatus::kSuccess)
 					|| (lockstatus == FileLock::LockStatus::kAlive));
 			return (lockstatus == FileLock::LockStatus::kAlive
@@ -1212,67 +988,71 @@ int main(int argc,char **argv) {
 		}
 	}
 
+
 #ifdef MFS_USE_MEMLOCK
 	if (lockmemory) {
 		if (getrlimit(RLIMIT_MEMLOCK,&rls)<0) {
-			mfs_errlog(LOG_WARNING,"error getting memory lock limits");
+			lzfs_pretty_errlog(LOG_WARNING,"error getting memory lock limits");
 		} else {
 			if (rls.rlim_cur!=RLIM_INFINITY && rls.rlim_max==RLIM_INFINITY) {
 				rls.rlim_cur = RLIM_INFINITY;
 				rls.rlim_max = RLIM_INFINITY;
 				if (setrlimit(RLIMIT_MEMLOCK,&rls)<0) {
-					mfs_errlog(LOG_WARNING,"error setting memory lock limit to unlimited");
+					lzfs_pretty_errlog(LOG_WARNING,"error setting memory lock limit to unlimited");
 				}
 			}
 			if (getrlimit(RLIMIT_MEMLOCK,&rls)<0) {
-				mfs_errlog(LOG_WARNING,"error getting memory lock limits");
+				lzfs_pretty_errlog(LOG_WARNING,"error getting memory lock limits");
 			} else {
 				if (rls.rlim_cur!=RLIM_INFINITY) {
-					mfs_errlog(LOG_WARNING,"can't set memory lock limit to unlimited");
+					lzfs_pretty_errlog(LOG_WARNING,"can't set memory lock limit to unlimited");
 				} else {
 					if (mlockall(MCL_CURRENT|MCL_FUTURE)<0) {
-						mfs_errlog(LOG_WARNING,"memory lock error");
+						lzfs_pretty_errlog(LOG_WARNING,"memory lock error");
 					} else {
-						mfs_syslog(LOG_NOTICE,"process memory was successfully locked in RAM");
+						lzfs_pretty_syslog(LOG_NOTICE,"process memory was successfully locked in RAM");
 					}
-			}       }
+				}
+			}
 		}
 	}
 #else
 	if (lockmemory) {
-		mfs_syslog(LOG_WARNING,"memory lock not supported !!!");
+		lzfs_pretty_syslog(LOG_WARNING,"trying to lock memory, but memory lock not supported");
 	}
 #endif
-	fprintf(stderr, "initializing %s modules ...\n", logappname.c_str());
 
 	if (initialize()) {
 		if (getrlimit(RLIMIT_NOFILE,&rls)==0) {
-			syslog(LOG_NOTICE,"open files limit: %lu",(unsigned long)(rls.rlim_cur));
+			lzfs_pretty_syslog(LOG_NOTICE,"open files limit: %lu",(unsigned long)(rls.rlim_cur));
 		}
-		fprintf(stderr, "%s daemon initialized properly\n", logappname.c_str());
+		lzfs_pretty_syslog(LOG_NOTICE, STR(APPNAME)" daemon initialized properly");
+#ifdef LIZARDFS_HAVE_SYSTEMD_SD_DAEMON_H
+		sd_notify(0, "READY=1\nSTATUS=" STR(APPNAME) " daemon initialized properly.\n");
+#endif
+		set_syslog_ident();
 		if (gRunAsDaemon) {
 			close_msg_channel();
 		}
 		if (initialize_late()) {
-			main_reloadregister(main_reload); // this will be the first thing to do
-			mainloop();
+			eventloop_reloadregister(main_reload); // this will be the first thing to do
+			eventloop_run();
 			ch=LIZARDFS_EXIT_STATUS_SUCCESS;
 		} else {
 			ch=LIZARDFS_EXIT_STATUS_ERROR;
 		}
 	} else {
-		fprintf(stderr,"error occured during initialization - exiting\n");
 		if (gRunAsDaemon) {
 			fputc(0,stderr);
 			close_msg_channel();
 		}
 		ch=LIZARDFS_EXIT_STATUS_ERROR;
 	}
-	destruct();
-	free_all_registered_entries();
+	eventloop_destruct();
+	eventloop_release_resources();
 	signal_cleanup();
 	cfg_term();
-	strerr_term();
 	closelog();
+	close_pam_session();
 	return ch;
 }
