@@ -1,5 +1,5 @@
 /*
-   Copyright 2005-2010 Jakub Kruszona-Zawadzki, Gemius SA, 2013 Skytechnology sp. z o.o..
+   Copyright 2005-2010 Jakub Kruszona-Zawadzki, Gemius SA, 2013-2014 EditShare, 2013-2015 Skytechnology sp. z o.o..
 
    This file was part of MooseFS and is part of LizardFS.
 
@@ -31,12 +31,15 @@
 #include <unistd.h>
 
 #include "common/cfg.h"
+#include "common/cwrap.h"
 #include "common/datapack.h"
+#include "common/exceptions.h"
+#include "common/event_loop.h"
 #include "common/goal.h"
 #include "common/main.h"
 #include "common/massert.h"
 #include "common/md5.h"
-#include "common/MFSCommunication.h"
+#include "protocol/MFSCommunication.h"
 #include "common/slogger.h"
 
 typedef struct _exports {
@@ -241,12 +244,12 @@ uint8_t exports_check(uint32_t ip,uint32_t version,uint8_t meta,const uint8_t *p
 	if (f==NULL) {
 		if (nopass) {
 			if (rndstate==0 || rndcode==NULL || passcode==NULL) {
-				return ERROR_NOPASSWORD;
+				return LIZARDFS_ERROR_NOPASSWORD;
 			} else {
-				return ERROR_BADPASSWORD;
+				return LIZARDFS_ERROR_BADPASSWORD;
 			}
 		}
-		return ERROR_EACCES;
+		return LIZARDFS_ERROR_EACCES;
 	}
 	*sesflags = f->sesflags;
 	*rootuid = f->rootuid;
@@ -257,7 +260,7 @@ uint8_t exports_check(uint32_t ip,uint32_t version,uint8_t meta,const uint8_t *p
 	*maxgoal = f->maxgoal;
 	*mintrashtime = f->mintrashtime;
 	*maxtrashtime = f->maxtrashtime;
-	return STATUS_OK;
+	return LIZARDFS_STATUS_OK;
 }
 
 void exports_freelist(exports *arec) {
@@ -418,7 +421,7 @@ int exports_parsegoal(char *goalstr,uint8_t *goal) {
 		// not the whole string was used by strtol
 		return -1;
 	}
-	if (value > UINT8_MAX || !goal::isGoalValid(value)) {
+	if (value > UINT8_MAX || !GoalId::isValid(value)) {
 		return -1;
 	}
 	*goal = value;
@@ -590,7 +593,7 @@ int exports_parseuidgid(char *maproot,uint32_t lineno,uint32_t *ruid,uint32_t *r
 		if (*eptr!=0) { // not only digits - treat it as a groupname
 			getgrnam_r(gptr+1,&grp,pwgrbuff,16384,&grrec);
 			if (grrec==NULL) {
-				mfs_arg_syslog(LOG_WARNING,"mfsexports/maproot: can't find group named '%s' defined in line: %" PRIu32,gptr+1,lineno);
+				lzfs_pretty_syslog(LOG_WARNING,"mfsexports/maproot: can't find group named '%s' defined in line: %" PRIu32,gptr+1,lineno);
 				return -1;
 			}
 			gid = grrec->gr_gid;
@@ -612,7 +615,7 @@ int exports_parseuidgid(char *maproot,uint32_t lineno,uint32_t *ruid,uint32_t *r
 	if (*eptr!=0) { // not only digits - treat it as a username
 		getpwnam_r(uptr,&pwd,pwgrbuff,16384,&pwrec);
 		if (pwrec==NULL) {
-			mfs_arg_syslog(LOG_WARNING,"mfsexports/maproot: can't find user named '%s' defined in line: %" PRIu32,uptr,lineno);
+			lzfs_pretty_syslog(LOG_WARNING,"mfsexports/maproot: can't find user named '%s' defined in line: %" PRIu32,uptr,lineno);
 			return -1;
 		}
 		*ruid = pwrec->pw_uid;
@@ -629,7 +632,7 @@ int exports_parseuidgid(char *maproot,uint32_t lineno,uint32_t *ruid,uint32_t *r
 	} else {
 		getpwuid_r(uid,&pwd,pwgrbuff,16384,&pwrec);
 		if (pwrec==NULL) {
-			mfs_arg_syslog(LOG_WARNING,"mfsexports/maproot: can't determine gid, because can't find user with uid %" PRIu32 " defined in line: %" PRIu32,uid,lineno);
+			lzfs_pretty_syslog(LOG_WARNING,"mfsexports/maproot: can't determine gid, because can't find user with uid %" PRIu32 " defined in line: %" PRIu32,uid,lineno);
 			return -1;
 		}
 		*ruid = pwrec->pw_uid;
@@ -666,7 +669,7 @@ int exports_parseoptions(char *opts,uint32_t lineno,exports *arec) {
 		case 'i':
 			if (strcmp(p,"ignoregid")==0) {
 				if (arec->meta) {
-					mfs_arg_syslog(LOG_WARNING,"meta option ignored: %s",p);
+					lzfs_pretty_syslog(LOG_WARNING,"meta option ignored: %s",p);
 				} else {
 					arec->sesflags |= SESFLAG_IGNOREGID;
 				}
@@ -676,14 +679,14 @@ int exports_parseoptions(char *opts,uint32_t lineno,exports *arec) {
 		case 'a':
 			if (strcmp(p,"allcanchangequota")==0) {
 				if (arec->meta) {
-					mfs_arg_syslog(LOG_WARNING,"meta option ignored: %s",p);
+					lzfs_pretty_syslog(LOG_WARNING,"meta option ignored: %s",p);
 				} else {
 					arec->sesflags |= SESFLAG_ALLCANCHANGEQUOTA;
 				}
 				o=1;
 			} else if (strcmp(p,"alldirs")==0) {
 				if (arec->meta) {
-					mfs_arg_syslog(LOG_WARNING,"meta option ignored: %s",p);
+					lzfs_pretty_syslog(LOG_WARNING,"meta option ignored: %s",p);
 				} else {
 					arec->alldirs = 1;
 				}
@@ -699,7 +702,7 @@ int exports_parseoptions(char *opts,uint32_t lineno,exports *arec) {
 		case 'n':
 			if (strcmp(p,"nomasterpermcheck")==0) {
 				if (arec->meta) {
-					mfs_arg_syslog(LOG_WARNING,"meta option ignored: %s",p);
+					lzfs_pretty_syslog(LOG_WARNING,"meta option ignored: %s",p);
 				} else {
 					arec->sesflags |= SESFLAG_NOMASTERPERMCHECK;
 				}
@@ -708,7 +711,7 @@ int exports_parseoptions(char *opts,uint32_t lineno,exports *arec) {
 				if (arec->meta) {
 					arec->sesflags |= SESFLAG_NONROOTMETA;
 				} else {
-					mfs_arg_syslog(LOG_WARNING,"non-meta option ignored: %s",p);
+					lzfs_pretty_syslog(LOG_WARNING,"non-meta option ignored: %s",p);
 				}
 				o=1;
 			}
@@ -717,7 +720,7 @@ int exports_parseoptions(char *opts,uint32_t lineno,exports *arec) {
 			if (strncmp(p,"maproot=",8)==0) {
 				o=1;
 				if (arec->meta) {
-					mfs_arg_syslog(LOG_WARNING,"meta option ignored: %s",p);
+					lzfs_pretty_syslog(LOG_WARNING,"meta option ignored: %s",p);
 				} else {
 					if (exports_parseuidgid(p+8,lineno,&arec->rootuid,&arec->rootgid)<0) {
 						return -1;
@@ -727,7 +730,7 @@ int exports_parseoptions(char *opts,uint32_t lineno,exports *arec) {
 			} else if (strncmp(p,"mapall=",7)==0) {
 				o=1;
 				if (arec->meta) {
-					mfs_arg_syslog(LOG_WARNING,"meta option ignored: %s",p);
+					lzfs_pretty_syslog(LOG_WARNING,"meta option ignored: %s",p);
 				} else {
 					if (exports_parseuidgid(p+7,lineno,&arec->mapalluid,&arec->mapallgid)<0) {
 						return -1;
@@ -764,53 +767,53 @@ int exports_parseoptions(char *opts,uint32_t lineno,exports *arec) {
 					}
 					arec->needpassword=1;
 				} else {
-					mfs_arg_syslog(LOG_WARNING,"mfsexports: incorrect md5pass definition (%s) in line: %" PRIu32,p,lineno);
+					lzfs_pretty_syslog(LOG_WARNING,"mfsexports: incorrect md5pass definition (%s) in line: %" PRIu32,p,lineno);
 					return -1;
 				}
 			} else if (strncmp(p,"minversion=",11)==0) {
 				o=1;
 				if (exports_parseversion(p+11,&arec->minversion)<0) {
-					mfs_arg_syslog(LOG_WARNING,"mfsexports: incorrect minversion definition (%s) in line: %" PRIu32,p,lineno);
+					lzfs_pretty_syslog(LOG_WARNING,"mfsexports: incorrect minversion definition (%s) in line: %" PRIu32,p,lineno);
 					return -1;
 				}
 			} else if (strncmp(p,"mingoal=",8)==0) {
 				o=1;
 				if (exports_parsegoal(p+8,&arec->mingoal)<0) {
-					mfs_arg_syslog(LOG_WARNING,"mfsexports: incorrect mingoal definition (%s) in line: %" PRIu32,p,lineno);
+					lzfs_pretty_syslog(LOG_WARNING,"mfsexports: incorrect mingoal definition (%s) in line: %" PRIu32,p,lineno);
 					return -1;
 				}
 				if (arec->mingoal>arec->maxgoal) {
-					mfs_arg_syslog(LOG_WARNING,"mfsexports: mingoal>maxgoal in definition (%s) in line: %" PRIu32,p,lineno);
+					lzfs_pretty_syslog(LOG_WARNING,"mfsexports: mingoal>maxgoal in definition (%s) in line: %" PRIu32,p,lineno);
 					return -1;
 				}
 			} else if (strncmp(p,"maxgoal=",8)==0) {
 				o=1;
 				if (exports_parsegoal(p+8,&arec->maxgoal)<0) {
-					mfs_arg_syslog(LOG_WARNING,"mfsexports: incorrect maxgoal definition (%s) in line: %" PRIu32,p,lineno);
+					lzfs_pretty_syslog(LOG_WARNING,"mfsexports: incorrect maxgoal definition (%s) in line: %" PRIu32,p,lineno);
 					return -1;
 				}
 				if (arec->mingoal>arec->maxgoal) {
-					mfs_arg_syslog(LOG_WARNING,"mfsexports: maxgoal<mingoal in definition (%s) in line: %" PRIu32,p,lineno);
+					lzfs_pretty_syslog(LOG_WARNING,"mfsexports: maxgoal<mingoal in definition (%s) in line: %" PRIu32,p,lineno);
 					return -1;
 				}
 			} else if (strncmp(p,"mintrashtime=",13)==0) {
 				o=1;
 				if (exports_parsetime(p+13,&arec->mintrashtime)<0) {
-					mfs_arg_syslog(LOG_WARNING,"mfsexports: incorrect mintrashtime definition (%s) in line: %" PRIu32,p,lineno);
+					lzfs_pretty_syslog(LOG_WARNING,"mfsexports: incorrect mintrashtime definition (%s) in line: %" PRIu32,p,lineno);
 					return -1;
 				}
 				if (arec->mintrashtime>arec->maxtrashtime) {
-					mfs_arg_syslog(LOG_WARNING,"mfsexports: mintrashtime>maxtrashtime in definition (%s) in line: %" PRIu32,p,lineno);
+					lzfs_pretty_syslog(LOG_WARNING,"mfsexports: mintrashtime>maxtrashtime in definition (%s) in line: %" PRIu32,p,lineno);
 					return -1;
 				}
 			} else if (strncmp(p,"maxtrashtime=",13)==0) {
 				o=1;
 				if (exports_parsetime(p+13,&arec->maxtrashtime)<0) {
-					mfs_arg_syslog(LOG_WARNING,"mfsexports: incorrect maxtrashtime definition (%s) in line: %" PRIu32,p,lineno);
+					lzfs_pretty_syslog(LOG_WARNING,"mfsexports: incorrect maxtrashtime definition (%s) in line: %" PRIu32,p,lineno);
 					return -1;
 				}
 				if (arec->mintrashtime>arec->maxtrashtime) {
-					mfs_arg_syslog(LOG_WARNING,"mfsexports: maxtrashtime<mintrashtime in definition (%s) in line: %" PRIu32,p,lineno);
+					lzfs_pretty_syslog(LOG_WARNING,"mfsexports: maxtrashtime<mintrashtime in definition (%s) in line: %" PRIu32,p,lineno);
 					return -1;
 				}
 			}
@@ -826,7 +829,7 @@ int exports_parseoptions(char *opts,uint32_t lineno,exports *arec) {
 			break;
 		}
 		if (o==0) {
-			mfs_arg_syslog(LOG_WARNING,"mfsexports: unknown option '%s' in line: %" PRIu32 " (ignored)",p,lineno);
+			lzfs_pretty_syslog(LOG_WARNING,"mfsexports: unknown option '%s' in line: %" PRIu32 " (ignored)",p,lineno);
 		}
 	}
 	return 0;
@@ -847,8 +850,8 @@ int exports_parseline(char *line,uint32_t lineno,exports *arec) {
 	arec->meta = 0;
 	arec->rootredefined = 0;
 	arec->sesflags = SESFLAG_READONLY;
-	arec->mingoal = goal::kMinGoal;
-	arec->maxgoal = goal::kMaxGoal;
+	arec->mingoal = GoalId::kMin;
+	arec->maxgoal = GoalId::kMax;
 	arec->mintrashtime = 0;
 	arec->maxtrashtime = UINT32_C(0xFFFFFFFF);
 	arec->rootuid = 999;
@@ -866,13 +869,13 @@ int exports_parseline(char *line,uint32_t lineno,exports *arec) {
 		p++;
 	}
 	if (*p==0) {
-		mfs_arg_syslog(LOG_WARNING,"mfsexports: incomplete definition in line: %" PRIu32,lineno);
+		lzfs_pretty_syslog(LOG_WARNING,"mfsexports: incomplete definition in line: %" PRIu32,lineno);
 		return -1;
 	}
 	*p=0;
 	p++;
 	if (exports_parsenet(net,&arec->fromip,&arec->toip)<0) {
-		mfs_arg_syslog(LOG_WARNING,"mfsexports: incorrect ip/network definition in line: %" PRIu32,lineno);
+		lzfs_pretty_syslog(LOG_WARNING,"mfsexports: incorrect ip/network definition in line: %" PRIu32,lineno);
 		return -1;
 	}
 
@@ -949,20 +952,15 @@ void exports_loadexports(void) {
 	fd = fopen(ExportsFileName,"r");
 	if (fd==NULL) {
 		if (errno==ENOENT) {
-			if (exports_records) {
-				syslog(LOG_WARNING,"mfsexports configuration file (%s) not found - exports not changed",ExportsFileName);
-			} else {
-				syslog(LOG_WARNING,"mfsexports configuration file (%s) not found - no exports !!!",ExportsFileName);
-			}
-			fprintf(stderr,"mfsexports configuration file (%s) not found - please create one (you can copy %s.dist to get a base configuration)\n",ExportsFileName,ExportsFileName);
+			throw InitializeException(
+					std::string("exports configuration file (") + ExportsFileName + ") not found"
+					" - please create one (you can copy " ETC_PATH "/mfsexports.cfg.dist"
+					" to get a base configuration)");
 		} else {
-			if (exports_records) {
-				mfs_arg_errlog(LOG_WARNING,"can't open mfsexports configuration file (%s) - exports not changed, error",ExportsFileName);
-			} else {
-				mfs_arg_errlog(LOG_WARNING,"can't open mfsexports configuration file (%s) - no exports !!!, error",ExportsFileName);
-			}
+			throw InitializeException(
+					std::string("can't open exports configuration file (") + ExportsFileName + "):"
+					+ errorString(errno));
 		}
-		return;
 	}
 	newexports = NULL;
 	netail = &newexports;
@@ -991,41 +989,32 @@ void exports_loadexports(void) {
 	free(arec);
 	if (ferror(fd)) {
 		fclose(fd);
-		syslog(LOG_WARNING,"error reading mfsexports file - exports not changed");
 		exports_freelist(newexports);
-		fprintf(stderr,"error reading mfsexports file - using defaults\n");
-		return;
+		throw InitializeException(
+				std::string("can't read exports configuration file (") + ExportsFileName + ")");
 	}
 	fclose(fd);
 	exports_freelist(exports_records);
 	exports_records = newexports;
-	mfs_syslog(LOG_NOTICE,"exports file has been loaded");
+	lzfs_pretty_syslog(LOG_INFO,"initialized exports from file %s", ExportsFileName);
 }
 
-void exports_reload(void) {
-	int fd;
+void exports_load(void) {
 	if (ExportsFileName) {
 		free(ExportsFileName);
 	}
-	if (!cfg_isdefined("EXPORTS_FILENAME")) {
-		ExportsFileName = strdup(ETC_PATH "/mfs/mfsexports.cfg");
-		passert(ExportsFileName);
-		if ((fd = open(ExportsFileName,O_RDONLY))<0 && errno==ENOENT) {
-			free(ExportsFileName);
-			ExportsFileName = strdup(ETC_PATH "/mfsexports.cfg");
-			if ((fd = open(ExportsFileName,O_RDONLY))>=0) {
-				mfs_syslog(LOG_WARNING,"default sysconf path has changed - please move mfsexports.cfg from " ETC_PATH "/ to " ETC_PATH "/mfs/");
-			}
-		}
-		if (fd>=0) {
-			close(fd);
-		}
-	} else {
-		ExportsFileName = cfg_getstr("EXPORTS_FILENAME", ETC_PATH "/mfs/mfsexports.cfg");
-	}
+	ExportsFileName = cfg_getstr("EXPORTS_FILENAME", ETC_PATH "/mfsexports.cfg");
 	exports_loadexports();
 }
 
+void exports_reload(void) {
+	try {
+		exports_load();
+	} catch (Exception& ex) {
+		lzfs_pretty_syslog(LOG_WARNING, "exports not changed because %s", ex.what());
+	}
+}
+
 void exports_term(void) {
 	exports_freelist(exports_records);
 	if (ExportsFileName) {
@@ -1036,12 +1025,12 @@ void exports_term(void) {
 int exports_init(void) {
 	exports_records = NULL;
 	ExportsFileName = NULL;
-	exports_reload();
-	if (exports_records==NULL) {
-		fprintf(stderr,"no exports defined !!!\n");
-		return -1;
+	exports_load();
+	if (exports_records == NULL) {
+		// File was empty
+		throw InitializeException(std::string("no exports defined in ") + ExportsFileName);
 	}
-	main_reloadregister(exports_reload);
-	main_destructregister(exports_term);
+	eventloop_reloadregister(exports_reload);
+	eventloop_destructregister(exports_term);
 	return 0;
 }
