@@ -1,5 +1,6 @@
+
 /*
-   Copyright 2005-2010 Jakub Kruszona-Zawadzki, Gemius SA, 2013 Skytechnology sp. z o.o..
+   Copyright 2005-2010 Jakub Kruszona-Zawadzki, Gemius SA, 2013-2014 EditShare, 2013-2015 Skytechnology sp. z o.o..
 
    This file was part of MooseFS and is part of LizardFS.
 
@@ -36,17 +37,18 @@
 #include "common/cfg.h"
 #include "common/crc.h"
 #include "common/datapack.h"
+#include "common/event_loop.h"
 #include "common/lizardfs_version.h"
-#include "common/main.h"
+#include "common/loop_watchdog.h"
 #include "common/massert.h"
-#include "common/matoml_communication.h"
 #include "common/metadata.h"
-#include "common/mltoma_communication.h"
-#include "common/MFSCommunication.h"
 #include "common/slogger.h"
 #include "common/sockets.h"
 #include "master/filesystem.h"
 #include "master/personality.h"
+#include "protocol/matoml.h"
+#include "protocol/MFSCommunication.h"
+#include "protocol/mltoma.h"
 
 #define MaxPacketSize 1500000
 #define OLD_CHANGES_BLOCK_SIZE 5000
@@ -75,6 +77,7 @@ typedef struct matomlserventry {
 	char *servstrip;                // human readable version of servip
 	uint32_t version;
 	uint32_t servip;
+	uint16_t servport;
 	bool shadow;
 
 	int metafd,chain1fd,chain2fd;
@@ -87,6 +90,12 @@ static int lsock;
 static int32_t lsockpdescpos;
 static bool gExiting = false;
 
+/// Miminal period (in seconds) between two metadata save processes requested by shadow masters
+static uint32_t gMinMetadataSaveRequestPeriod_s;
+
+/// Timestamp of the last metadata save request
+static uint32_t gLastMetadataSaveRequestTimestamp = 0;
+
 typedef struct old_changes_entry {
 	uint64_t version;
 	uint32_t length;
@@ -132,9 +141,7 @@ public:
 	 */
 	void handleRequests(uint8_t status) {
 		for (matomlserventry* eptr : shadowRequests_) {
-			std::vector<uint8_t> reply;
-			matoml::changelogApplyError::serialize(reply, status);
-			matomlserv_createpacket(eptr, std::move(reply));
+			matomlserv_createpacket(eptr, matoml::changelogApplyError::build(status));
 		}
 		shadowRequests_.clear();
 	}
@@ -147,7 +154,7 @@ private:
  *
  * \param status - status to be forwarded.
  */
-void matomlserv_broadcast_filesystem(uint8_t status) {
+void matomlserv_broadcast_metadata_saved(uint8_t status) {
 	gShadowQueue.handleRequests(status);
 }
 
@@ -182,7 +189,7 @@ void matomlserv_store_logstring(uint64_t version,uint8_t *logstr,uint32_t logstr
 	if (old_changes_current==NULL || old_changes_head==NULL || old_changes_current->entries>=OLD_CHANGES_BLOCK_SIZE) {
 		oc = (old_changes_block*) malloc(sizeof(old_changes_block));
 		passert(oc);
-		ts = main_time();
+		ts = eventloop_time();
 		oc->entries = 0;
 		oc->minversion = version;
 		oc->mintimestamp = ts;
@@ -214,7 +221,7 @@ uint32_t matomlserv_mloglist_size(void) {
 	uint32_t i;
 	i=0;
 	for (eptr = matomlservhead ; eptr ; eptr=eptr->next) {
-		if (eptr->mode!=KILL) {
+		if (!eptr->shadow && eptr->mode!=KILL) {
 			i++;
 		}
 	}
@@ -224,13 +231,33 @@ uint32_t matomlserv_mloglist_size(void) {
 void matomlserv_mloglist_data(uint8_t *ptr) {
 	matomlserventry *eptr;
 	for (eptr = matomlservhead ; eptr ; eptr=eptr->next) {
-		if (eptr->mode!=KILL) {
+		if (!eptr->shadow && eptr->mode!=KILL) {
 			put32bit(&ptr,eptr->version);
 			put32bit(&ptr,eptr->servip);
 		}
 	}
 }
 
+std::vector<MetadataserverListEntry> matomlserv_shadows() {
+	std::vector<MetadataserverListEntry> ret;
+	for (matomlserventry* eptr = matomlservhead; eptr; eptr=eptr->next) {
+		if (eptr->shadow) {
+			ret.emplace_back(eptr->servip, eptr->servport, eptr->version);
+		}
+	}
+	return ret;
+}
+
+uint32_t matomlserv_shadows_count() {
+	uint32_t count = 0;
+	for (matomlserventry* eptr = matomlservhead; eptr; eptr=eptr->next) {
+		if (eptr->shadow) {
+			count++;
+		}
+	}
+	return count;
+}
+
 void matomlserv_status(void) {
 	matomlserventry *eptr;
 	for (eptr = matomlservhead ; eptr ; eptr=eptr->next) {
@@ -238,7 +265,7 @@ void matomlserv_status(void) {
 			return;
 		}
 	}
-	syslog(LOG_WARNING,"no meta loggers connected !!!");
+	lzfs_pretty_syslog(LOG_WARNING,"no meta loggers connected !!!");
 }
 
 char* matomlserv_makestrip(uint32_t ip) {
@@ -306,11 +333,10 @@ void matomlserv_send_old_changes(matomlserventry *eptr,uint64_t version) {
 	uint8_t start=0;
 	uint32_t i;
 	if (old_changes_head==NULL) {
-		// syslog(LOG_WARNING,"meta logger wants old changes, but storage is disabled");
 		return;
 	}
 	if (old_changes_head->minversion>version) {
-		syslog(LOG_WARNING,"meta logger wants changes since version: %" PRIu64 ", but minimal version in storage is: %" PRIu64,version,old_changes_head->minversion);
+		lzfs_pretty_syslog(LOG_WARNING,"meta logger wants changes since version: %" PRIu64 ", but minimal version in storage is: %" PRIu64,version,old_changes_head->minversion);
 		// TODO(msulikowski) send a special message which will cause the shadow master to unload fs
 	}
 	for (oc=old_changes_head ; oc ; oc=oc->next) {
@@ -335,24 +361,24 @@ void matomlserv_send_old_changes(matomlserventry *eptr,uint64_t version) {
 
 void matomlserv_register(matomlserventry *eptr,const uint8_t *data,uint32_t length) {
 	if (eptr->version>0) {
-		syslog(LOG_WARNING,"got register message from registered metalogger !!!");
+		lzfs_pretty_syslog(LOG_WARNING,"got register message from registered metalogger !!!");
 		eptr->mode=KILL;
 		return;
 	}
 	if (length<1) {
-		syslog(LOG_NOTICE,"MLTOMA_REGISTER - wrong size (%" PRIu32 ")",length);
+		lzfs_pretty_syslog(LOG_NOTICE,"MLTOMA_REGISTER - wrong size (%" PRIu32 ")",length);
 		eptr->mode=KILL;
 		return;
 	} else {
 		uint8_t rversion = get8bit(&data);
 		if (rversion < 1 || rversion > 4) {
-			syslog(LOG_NOTICE,"MLTOMA_REGISTER - wrong version (%" PRIu8 ")",rversion);
+			lzfs_pretty_syslog(LOG_NOTICE,"MLTOMA_REGISTER - wrong version (%" PRIu8 ")",rversion);
 			eptr->mode=KILL;
 			return;
 		}
 		static const uint32_t expected_length[] = {0, 7, 7+8, 7, 7+8};
 		if (length != expected_length[rversion]) {
-			syslog(LOG_NOTICE,"MLTOMA_REGISTER (ver %" PRIu8 ") - wrong size (%" PRIu32
+			lzfs_pretty_syslog(LOG_NOTICE,"MLTOMA_REGISTER (ver %" PRIu8 ") - wrong size (%" PRIu32
 					"/%" PRIu32 ")", rversion, length, expected_length[rversion]);
 			eptr->mode=KILL;
 			return;
@@ -360,9 +386,10 @@ void matomlserv_register(matomlserventry *eptr,const uint8_t *data,uint32_t leng
 		eptr->version = get32bit(&data);
 		eptr->timeout = get16bit(&data);
 		eptr->shadow = (rversion == 3 || rversion == 4);
-		if (eptr->version < LIZARDFS_VERSHEX) {
-			syslog(LOG_NOTICE,
-					"MLTOMA_REGISTER (ver %" PRIu8 ") - rejected old client (v%s) from %s",
+		if (eptr->shadow) {
+		// supported shadow master servers register using LIZ_MLTOMA_REGISTER_SHADOW packet
+			lzfs_pretty_syslog(LOG_NOTICE,
+					"MLTOMA_REGISTER (ver %" PRIu8 ") - rejected old shadow master (v%s) from %s",
 					rversion, lizardfsVersionToString(eptr->version).c_str(), eptr->servstrip);
 			eptr->mode=KILL;
 			return;
@@ -372,7 +399,7 @@ void matomlserv_register(matomlserventry *eptr,const uint8_t *data,uint32_t leng
 			matomlserv_send_old_changes(eptr,minversion);
 		}
 		if (eptr->timeout<10) {
-			syslog(LOG_NOTICE,"MLTOMA_REGISTER communication timeout too small (%" PRIu16 " seconds - should be at least 10 seconds)",eptr->timeout);
+			lzfs_pretty_syslog(LOG_NOTICE,"MLTOMA_REGISTER communication timeout too small (%" PRIu16 " seconds - should be at least 10 seconds)",eptr->timeout);
 			if (eptr->timeout<3) {
 				eptr->timeout=3;
 			}
@@ -388,7 +415,7 @@ void matomlserv_register_shadow(matomlserventry *eptr, const uint8_t *data, uint
 	eptr->version = version;
 	eptr->shadow = true;
 	if (eptr->timeout < 10) {
-		syslog(LOG_NOTICE,
+		lzfs_pretty_syslog(LOG_NOTICE,
 				"MLTOMA_REGISTER_SHADOW communication timeout too small (%" PRIu32 " milliseconds)"
 				" - should be at least 10 seconds; increasing to 10 seconds", timeout_ms);
 		if (eptr->timeout < 10) {
@@ -396,12 +423,10 @@ void matomlserv_register_shadow(matomlserventry *eptr, const uint8_t *data, uint
 		}
 	}
 	if (eptr->version < LIZARDFS_VERSHEX) {
-		syslog(LOG_NOTICE,
+		lzfs_pretty_syslog(LOG_NOTICE,
 				"MLTOMA_REGISTER_SHADOW - rejected old client (v%s) from %s",
 				lizardfsVersionToString(eptr->version).c_str(), eptr->servstrip);
-		std::vector<uint8_t> reply;
-		matoml::registerShadow::serialize(reply, uint8_t(ERROR_REGISTER));
-		matomlserv_createpacket(eptr, std::move(reply));
+		matomlserv_createpacket(eptr, matoml::registerShadow::build(uint8_t(LIZARDFS_ERROR_REGISTER)));
 		return;
 	}
 
@@ -418,18 +443,24 @@ void matomlserv_register_shadow(matomlserventry *eptr, const uint8_t *data, uint
 		replyVersion = myMedatataVersion;
 	}
 
-	std::vector<uint8_t> reply;
-	matoml::registerShadow::serialize(reply, LIZARDFS_VERSHEX, replyVersion);
-	matomlserv_createpacket(eptr, std::move(reply));
+	matomlserv_createpacket(eptr, matoml::registerShadow::build(LIZARDFS_VERSHEX, replyVersion));
 	matomlserv_send_old_changes(eptr, replyVersion - 1); // this function expects lastlogversion
 }
 
+void matomlserv_matoclport(matomlserventry *eptr, const uint8_t *data, uint32_t length) {
+	mltoma::matoclport::deserialize(data, length, eptr->servport);
+}
+
 void matomlserv_download_start(matomlserventry *eptr,const uint8_t *data,uint32_t length) {
 	if (length!=1) {
-		syslog(LOG_NOTICE,"MLTOMA_DOWNLOAD_START - wrong size (%" PRIu32 "/1)",length);
+		lzfs_pretty_syslog(LOG_NOTICE,"MLTOMA_DOWNLOAD_START - wrong size (%" PRIu32 "/1)",length);
 		eptr->mode=KILL;
 		return;
 	}
+	if (gExiting) {
+		lzfs_pretty_syslog(LOG_NOTICE,"MLTOMA_DOWNLOAD_START - ignoring in the exit phase");
+		return;
+	}
 	uint8_t filenum = get8bit(&data);
 
 	if ((filenum == DOWNLOAD_METADATA_MFS) || (filenum == DOWNLOAD_SESSIONS_MFS)) {
@@ -493,12 +524,16 @@ void matomlserv_download_data(matomlserventry *eptr,const uint8_t *data,uint32_t
 	ssize_t ret;
 
 	if (length!=12) {
-		syslog(LOG_NOTICE,"MLTOMA_DOWNLOAD_DATA - wrong size (%" PRIu32 "/12)",length);
+		lzfs_pretty_syslog(LOG_NOTICE,"MLTOMA_DOWNLOAD_DATA - wrong size (%" PRIu32 "/12)",length);
 		eptr->mode=KILL;
 		return;
 	}
+	if (gExiting) {
+		lzfs_pretty_syslog(LOG_NOTICE,"MLTOMA_DOWNLOAD_DATA - ignoring in the exit phase");
+		return;
+	}
 	if (eptr->metafd<0) {
-		syslog(LOG_NOTICE,"MLTOMA_DOWNLOAD_DATA - file not opened");
+		lzfs_pretty_syslog(LOG_NOTICE,"MLTOMA_DOWNLOAD_DATA - file not opened");
 		eptr->mode=KILL;
 		return;
 	}
@@ -507,14 +542,9 @@ void matomlserv_download_data(matomlserventry *eptr,const uint8_t *data,uint32_t
 	ptr = matomlserv_createpacket(eptr,MATOML_DOWNLOAD_DATA,16+leng);
 	put64bit(&ptr,offset);
 	put32bit(&ptr,leng);
-#ifdef LIZARDFS_HAVE_PREAD
 	ret = pread(eptr->metafd,ptr+4,leng,offset);
-#else /* LIZARDFS_HAVE_PWRITE */
-	lseek(eptr->metafd,offset,SEEK_SET);
-	ret = read(eptr->metafd,ptr+4,leng);
-#endif /* LIZARDFS_HAVE_PWRITE */
 	if (ret!=(ssize_t)leng) {
-		mfs_errlog_silent(LOG_NOTICE,"error reading metafile");
+		lzfs_silent_errlog(LOG_NOTICE,"error reading metafile");
 		eptr->mode=KILL;
 		return;
 	}
@@ -525,7 +555,7 @@ void matomlserv_download_data(matomlserventry *eptr,const uint8_t *data,uint32_t
 void matomlserv_download_end(matomlserventry *eptr,const uint8_t *data,uint32_t length) {
 	(void)data;
 	if (length!=0) {
-		syslog(LOG_NOTICE,"MLTOMA_DOWNLOAD_END - wrong size (%" PRIu32 "/0)",length);
+		lzfs_pretty_syslog(LOG_NOTICE,"MLTOMA_DOWNLOAD_END - wrong size (%" PRIu32 "/0)",length);
 		eptr->mode=KILL;
 		return;
 	}
@@ -538,13 +568,31 @@ void matomlserv_download_end(matomlserventry *eptr,const uint8_t *data,uint32_t
 void matomlserv_changelog_apply_error(matomlserventry *eptr, const uint8_t *data, uint32_t length) {
 	uint8_t recvStatus;
 	mltoma::changelogApplyError::deserialize(data, length, recvStatus);
-	DEBUG_LOG("master.mltoma_changelog_apply_error") << "status: " << int(recvStatus);
-	syslog(LOG_INFO, "LIZ_MLTOMA_CHANGELOG_APPLY_ERROR, status: %s - dumping metadata",
-			mfsstrerr(recvStatus));
-	gShadowQueue.addRequest(eptr);
-	fs_storeall(MetadataDumper::kBackgroundDump);
-	if (recvStatus == ERROR_BADMETADATACHECKSUM) {
-		fs_start_checksum_recalculation();
+	if (gExiting) {
+		lzfs_pretty_syslog(LOG_NOTICE,"LIZ_MLTOMA_CHANGELOG_APPLY_ERROR - ignoring in the exit phase");
+		return;
+	}
+
+	int32_t secondsSinceLastRequest = eventloop_time() - gLastMetadataSaveRequestTimestamp;
+	if (secondsSinceLastRequest >= int32_t(gMinMetadataSaveRequestPeriod_s)) {
+		gLastMetadataSaveRequestTimestamp = eventloop_time();
+		lzfs_silent_syslog(LOG_DEBUG, "master.mltoma_changelog_apply_error: do");
+		lzfs_pretty_syslog(LOG_INFO,
+				"LIZ_MLTOMA_CHANGELOG_APPLY_ERROR, status: %s - storing metadata",
+				lizardfs_error_string(recvStatus));
+		gShadowQueue.addRequest(eptr);
+		fs_storeall(MetadataDumper::kBackgroundDump);
+		if (recvStatus == LIZARDFS_ERROR_BADMETADATACHECKSUM) {
+			fs_start_checksum_recalculation();
+		}
+	} else {
+		lzfs_silent_syslog(LOG_DEBUG, "master.mltoma_changelog_apply_error: delay");
+		lzfs_pretty_syslog(LOG_INFO,
+				"LIZ_MLTOMA_CHANGELOG_APPLY_ERROR, status: %s - "
+				"refusing to store metadata because only %" PRIi32 " seconds elapsed since the "
+				"previous request and METADATA_SAVE_REQUEST_MIN_PERIOD=%" PRIu32,
+				lizardfs_error_string(recvStatus), secondsSinceLastRequest, gMinMetadataSaveRequestPeriod_s);
+		matomlserv_createpacket(eptr, matoml::changelogApplyError::build(LIZARDFS_ERROR_DELAYED));
 	}
 }
 
@@ -619,12 +667,15 @@ void matomlserv_gotpacket(matomlserventry *eptr,uint32_t type,const uint8_t *dat
 			case LIZ_MLTOMA_CHANGELOG_APPLY_ERROR:
 				matomlserv_changelog_apply_error(eptr, data, length);
 				break;
+			case LIZ_MLTOMA_CLTOMA_PORT:
+				matomlserv_matoclport(eptr, data, length);
+				break;
 			default:
-				syslog(LOG_NOTICE,"master <-> metaloggers module: got unknown message (type:%" PRIu32 ")",type);
+				lzfs_pretty_syslog(LOG_NOTICE,"master <-> metaloggers module: got unknown message (type:%" PRIu32 ")",type);
 				eptr->mode=KILL;
 		}
 	} catch (IncorrectDeserializationException& ex) {
-		syslog(LOG_NOTICE, "Packet 0x%" PRIX32 " - cannot deserialize: %s", type, ex.what());
+		lzfs_pretty_syslog(LOG_NOTICE, "Packet 0x%" PRIX32 " - can't deserialize: %s", type, ex.what());
 		eptr->mode = KILL;
 	}
 }
@@ -632,7 +683,7 @@ void matomlserv_gotpacket(matomlserventry *eptr,uint32_t type,const uint8_t *dat
 void matomlserv_term(void) {
 	matomlserventry *eptr,*eaptr;
 	packetstruct *pptr,*paptr;
-	syslog(LOG_INFO,"master <-> metaloggers module: closing %s:%s",ListenHost,ListenPort);
+	lzfs_pretty_syslog(LOG_INFO,"master <-> metaloggers module: closing %s:%s",ListenHost,ListenPort);
 	tcpclose(lsock);
 
 	eptr = matomlservhead;
@@ -654,6 +705,7 @@ void matomlserv_term(void) {
 		}
 		eaptr = eptr;
 		eptr = eptr->next;
+		gShadowQueue.removeRequest(eaptr);
 		free(eaptr);
 	}
 	matomlservhead=NULL;
@@ -663,19 +715,22 @@ void matomlserv_term(void) {
 }
 
 void matomlserv_read(matomlserventry *eptr) {
+	SignalLoopWatchdog watchdog;
 	int32_t i;
 	uint32_t type,size;
 	const uint8_t *ptr;
-	for (;;) {
+
+	watchdog.start();
+	while (eptr->mode != KILL) {
 		i=read(eptr->sock,eptr->inputpacket.startptr,eptr->inputpacket.bytesleft);
 		if (i==0) {
-			syslog(LOG_NOTICE,"connection with ML(%s) has been closed by peer",eptr->servstrip);
+			lzfs_pretty_syslog(LOG_NOTICE,"connection with ML(%s) has been closed by peer",eptr->servstrip);
 			eptr->mode = KILL;
 			return;
 		}
 		if (i<0) {
 			if (errno!=EAGAIN) {
-				mfs_arg_errlog_silent(LOG_NOTICE,"read from ML(%s) error",eptr->servstrip);
+				lzfs_silent_errlog(LOG_NOTICE,"read from ML(%s) error",eptr->servstrip);
 				eptr->mode = KILL;
 			}
 			return;
@@ -693,7 +748,7 @@ void matomlserv_read(matomlserventry *eptr) {
 
 			if (size>0) {
 				if (size>MaxPacketSize) {
-					syslog(LOG_WARNING,"ML(%s) packet too long (%" PRIu32 "/%u)",eptr->servstrip,size,MaxPacketSize);
+					lzfs_pretty_syslog(LOG_WARNING,"ML(%s) packet too long (%" PRIu32 "/%u)",eptr->servstrip,size,MaxPacketSize);
 					eptr->mode = KILL;
 					return;
 				}
@@ -722,13 +777,21 @@ void matomlserv_read(matomlserventry *eptr) {
 				free(eptr->inputpacket.packet);
 			}
 			eptr->inputpacket.packet=NULL;
+			break;
+		}
+
+		if (watchdog.expired()) {
+			break;
 		}
 	}
 }
 
 void matomlserv_write(matomlserventry *eptr) {
+	SignalLoopWatchdog watchdog;
 	packetstruct *pack;
 	int32_t i;
+
+	watchdog.start();
 	for (;;) {
 		pack = eptr->outputhead;
 		if (pack==NULL) {
@@ -737,7 +800,7 @@ void matomlserv_write(matomlserventry *eptr) {
 		i=write(eptr->sock,pack->startptr,pack->bytesleft);
 		if (i<0) {
 			if (errno!=EAGAIN) {
-				mfs_arg_errlog_silent(LOG_NOTICE,"write to ML(%s) error",eptr->servstrip);
+				lzfs_silent_errlog(LOG_NOTICE,"write to ML(%s) error",eptr->servstrip);
 				eptr->mode = KILL;
 			}
 			return;
@@ -753,34 +816,32 @@ void matomlserv_write(matomlserventry *eptr) {
 			eptr->outputtail = &(eptr->outputhead);
 		}
 		free(pack);
+
+		if (watchdog.expired()) {
+			break;
+		}
 	}
 }
 
-void matomlserv_desc(struct pollfd *pdesc,uint32_t *ndesc) {
-	uint32_t pos = *ndesc;
+void matomlserv_desc(std::vector<pollfd> &pdesc) {
 	matomlserventry *eptr;
 	if (!gExiting) {
-		pdesc[pos].fd = lsock;
-		pdesc[pos].events = POLLIN;
-		lsockpdescpos = pos;
-		pos++;
+		pdesc.push_back({lsock,POLLIN,0});
+		lsockpdescpos = pdesc.size() - 1;
 	} else {
 		lsockpdescpos = -1;
 	}
 	for (eptr=matomlservhead ; eptr ; eptr=eptr->next) {
-		pdesc[pos].fd = eptr->sock;
-		pdesc[pos].events = POLLIN;
-		eptr->pdescpos = pos;
+		pdesc.push_back({eptr->sock,POLLIN,0});
+		eptr->pdescpos = pdesc.size() - 1;
 		if (eptr->outputhead!=NULL) {
-			pdesc[pos].events |= POLLOUT;
+			pdesc.back().events |= POLLOUT;
 		}
-		pos++;
 	}
-	*ndesc = pos;
 }
 
-void matomlserv_serve(struct pollfd *pdesc) {
-	uint32_t now=main_time();
+void matomlserv_serve(const std::vector<pollfd> &pdesc) {
+	uint32_t now=eventloop_time();
 	matomlserventry *eptr,**kptr;
 	packetstruct *pptr,*paptr;
 	int ns;
@@ -788,7 +849,7 @@ void matomlserv_serve(struct pollfd *pdesc) {
 	if (lsockpdescpos>=0 && (pdesc[lsockpdescpos].revents & POLLIN)) {
 		ns=tcpaccept(lsock);
 		if (ns<0) {
-			mfs_errlog_silent(LOG_NOTICE,"Master<->ML socket: accept error");
+			lzfs_silent_errlog(LOG_NOTICE,"master<->ML socket: accept error");
 		} else if (metadataserver::isMaster()) {
 			tcpnonblock(ns);
 			tcpnodelay(ns);
@@ -808,6 +869,8 @@ void matomlserv_serve(struct pollfd *pdesc) {
 			eptr->outputhead = NULL;
 			eptr->outputtail = &(eptr->outputhead);
 			eptr->timeout = 10;
+			eptr->servport = 0;// For shadow masters this will be changed to their MATOCL_SERV_PORT
+			eptr->shadow = false;
 
 			tcpgetpeer(eptr->sock,&(eptr->servip),NULL);
 			eptr->servstrip = matomlserv_makestrip(eptr->servip);
@@ -836,7 +899,9 @@ void matomlserv_serve(struct pollfd *pdesc) {
 		if ((uint32_t)(eptr->lastread+eptr->timeout)<(uint32_t)now) {
 			eptr->mode = KILL;
 		}
-		if ((uint32_t)(eptr->lastwrite+(eptr->timeout/3))<(uint32_t)now && eptr->outputhead==NULL) {
+		if ((uint32_t)(eptr->lastwrite+(eptr->timeout/3))<(uint32_t)now
+				&& eptr->outputhead==NULL
+				&& !gExiting) {
 			matomlserv_createpacket(eptr,ANTOAN_NOP,0);
 		}
 	}
@@ -870,26 +935,35 @@ void matomlserv_serve(struct pollfd *pdesc) {
 
 void matomlserv_wantexit(void) {
 	gExiting = true;
+	for (matomlserventry *eptr = matomlservhead; eptr != nullptr; eptr = eptr->next) {
+		matomlserv_createpacket(eptr, matoml::endSession::build());
+	}
+	// Now we won't create any new packets, but we will wait for all existing packets to be
+	// transmitted to shadow masters and metaloggers
 }
 
 int matomlserv_canexit(void) {
-	for (matomlserventry *eptr = matomlservhead; eptr != nullptr; eptr = eptr->next) {
-		if (eptr->outputhead != nullptr) {
-			return 0;
-		}
-	}
-	return 1;
+	// Exit when all connections are closed
+	return (matomlservhead == nullptr);
 }
 
 void matomlserv_become_master() {
-	main_timeregister(TIMEMODE_SKIP_LATE,3600,0,matomlserv_status);
+	eventloop_timeregister(TIMEMODE_SKIP_LATE,3600,0,matomlserv_status);
 	return;
 }
 
+/// Read values from the config file into global variables.
+/// Used on init and reload.
+void matomlserv_read_config_file() {
+	gMinMetadataSaveRequestPeriod_s = cfg_getuint32("METADATA_SAVE_REQUEST_MIN_PERIOD", 1800);
+}
+
 void matomlserv_reload(void) {
 	char *oldListenHost,*oldListenPort;
 	int newlsock;
 
+	matomlserv_read_config_file();
+
 	oldListenHost = ListenHost;
 	oldListenPort = ListenPort;
 	ListenHost = cfg_getstr("MATOML_LISTEN_HOST","*");
@@ -897,13 +971,13 @@ void matomlserv_reload(void) {
 	if (strcmp(oldListenHost,ListenHost)==0 && strcmp(oldListenPort,ListenPort)==0) {
 		free(oldListenHost);
 		free(oldListenPort);
-		mfs_arg_syslog(LOG_NOTICE,"master <-> metaloggers module: socket address hasn't changed (%s:%s)",ListenHost,ListenPort);
+		lzfs_pretty_syslog(LOG_NOTICE,"master <-> metaloggers module: socket address hasn't changed (%s:%s)",ListenHost,ListenPort);
 		return;
 	}
 
 	newlsock = tcpsocket();
 	if (newlsock<0) {
-		mfs_errlog(LOG_WARNING,"master <-> metaloggers module: socket address has changed, but can't create new socket");
+		lzfs_pretty_errlog(LOG_WARNING,"master <-> metaloggers module: socket address has changed, but can't create new socket");
 		free(ListenHost);
 		free(ListenPort);
 		ListenHost = oldListenHost;
@@ -914,10 +988,10 @@ void matomlserv_reload(void) {
 	tcpnodelay(newlsock);
 	tcpreuseaddr(newlsock);
 	if (tcpsetacceptfilter(newlsock)<0 && errno!=ENOTSUP) {
-		mfs_errlog_silent(LOG_NOTICE,"master <-> metaloggers module: can't set accept filter");
+		lzfs_silent_errlog(LOG_NOTICE,"master <-> metaloggers module: can't set accept filter");
 	}
 	if (tcpstrlisten(newlsock,ListenHost,ListenPort,100)<0) {
-		mfs_arg_errlog(LOG_ERR,"master <-> metaloggers module: socket address has changed, but can't listen on socket (%s:%s)",ListenHost,ListenPort);
+		lzfs_pretty_errlog(LOG_ERR,"master <-> metaloggers module: socket address has changed, but can't listen on socket (%s:%s)",ListenHost,ListenPort);
 		free(ListenHost);
 		free(ListenPort);
 		ListenHost = oldListenHost;
@@ -925,7 +999,7 @@ void matomlserv_reload(void) {
 		tcpclose(newlsock);
 		return;
 	}
-	mfs_arg_syslog(LOG_NOTICE,"master <-> metaloggers module: socket address has changed, now listen on %s:%s",ListenHost,ListenPort);
+	lzfs_pretty_syslog(LOG_NOTICE,"master <-> metaloggers module: socket address has changed, now listen on %s:%s",ListenHost,ListenPort);
 	free(oldListenHost);
 	free(oldListenPort);
 	tcpclose(lsock);
@@ -933,46 +1007,46 @@ void matomlserv_reload(void) {
 
 	ChangelogSecondsToRemember = cfg_getuint16("MATOML_LOG_PRESERVE_SECONDS",600);
 	if (ChangelogSecondsToRemember>3600) {
-		syslog(LOG_WARNING,"Number of seconds of change logs to be preserved in master is too big (%" PRIu16 ") - decreasing to 3600 seconds",ChangelogSecondsToRemember);
+		lzfs_pretty_syslog(LOG_WARNING,"Number of seconds of change logs to be preserved in master is too big (%" PRIu16 ") - decreasing to 3600 seconds",ChangelogSecondsToRemember);
 		ChangelogSecondsToRemember=3600;
 	}
-	if (metadataserver::isDuringPersonalityChange()) {
-		matomlserv_become_master();
-	}
 }
 
 int matomlserv_init(void) {
+	matomlserv_read_config_file();
 	ListenHost = cfg_getstr("MATOML_LISTEN_HOST","*");
 	ListenPort = cfg_getstr("MATOML_LISTEN_PORT","9419");
 
 	lsock = tcpsocket();
 	if (lsock<0) {
-		mfs_errlog(LOG_ERR,"master <-> metaloggers module: can't create socket");
+		lzfs_pretty_errlog(LOG_ERR,"master <-> metaloggers module: can't create socket");
 		return -1;
 	}
 	tcpnonblock(lsock);
 	tcpnodelay(lsock);
 	tcpreuseaddr(lsock);
 	if (tcpsetacceptfilter(lsock)<0 && errno!=ENOTSUP) {
-		mfs_errlog_silent(LOG_NOTICE,"master <-> metaloggers module: can't set accept filter");
+		lzfs_silent_errlog(LOG_NOTICE,"master <-> metaloggers module: can't set accept filter");
 	}
+
 	if (tcpstrlisten(lsock,ListenHost,ListenPort,100)<0) {
-		mfs_arg_errlog(LOG_ERR,"master <-> metaloggers module: can't listen on %s:%s",ListenHost,ListenPort);
+		lzfs_pretty_errlog(LOG_ERR,"master <-> metaloggers module: can't listen on %s:%s",ListenHost,ListenPort);
 		return -1;
 	}
-	mfs_arg_syslog(LOG_NOTICE,"master <-> metaloggers module: listen on %s:%s",ListenHost,ListenPort);
+	lzfs_pretty_syslog(LOG_NOTICE,"master <-> metaloggers module: listen on %s:%s",ListenHost,ListenPort);
 
 	matomlservhead = NULL;
 	ChangelogSecondsToRemember = cfg_getuint16("MATOML_LOG_PRESERVE_SECONDS",600);
 	if (ChangelogSecondsToRemember>3600) {
-		syslog(LOG_WARNING,"Number of seconds of change logs to be preserved in master is too big (%" PRIu16 ") - decreasing to 3600 seconds",ChangelogSecondsToRemember);
+		lzfs_pretty_syslog(LOG_WARNING,"Number of seconds of change logs to be preserved in master is too big (%" PRIu16 ") - decreasing to 3600 seconds",ChangelogSecondsToRemember);
 		ChangelogSecondsToRemember=3600;
 	}
-	main_wantexitregister(matomlserv_wantexit);
-	main_canexitregister(matomlserv_canexit);
-	main_reloadregister(matomlserv_reload);
-	main_destructregister(matomlserv_term);
-	main_pollregister(matomlserv_desc,matomlserv_serve);
+	eventloop_wantexitregister(matomlserv_wantexit);
+	eventloop_canexitregister(matomlserv_canexit);
+	eventloop_reloadregister(matomlserv_reload);
+	metadataserver::registerFunctionCalledOnPromotion(matomlserv_become_master);
+	eventloop_destructregister(matomlserv_term);
+	eventloop_pollregister(matomlserv_desc,matomlserv_serve);
 	if (metadataserver::isMaster()) {
 		matomlserv_become_master();
 	}
