@@ -1,4 +1,4 @@
-timeout_set 1 hour
+timeout_set 2 hours
 assert_program_installed git
 assert_program_installed cmake
 assert_program_installed rsync
@@ -24,9 +24,9 @@ CHUNKSERVERS=3 \
 	setup_local_empty_lizardfs info
 
 cd "${info[mount0]}"
-for goal in 1 2 3; do
+for goal in 1 2 3 xor2; do
 	mkdir "goal_$goal"
-	mfssetgoal "$goal" "goal_$goal"
+	lizardfs setgoal "$goal" "goal_$goal"
 	test_worker "goal_$goal" &
 done
 wait
