@@ -1,3 +1,5 @@
+timeout_set 1 minute
+
 verify_file() {
 	local file="$1"
 	local grep='grep -Hn --color=auto'
@@ -32,5 +34,8 @@ fi
 git ls-tree -r --name-only HEAD \
 		| egrep '[.](cmake|txt|cc|c|h|sh|inc|in|cfg)$' \
 		| grep -v 'mfs[.]cgi[.]in' \
+		| grep -v 'lizardfs_c_api[.]h' \
+		| grep -v 'lizardfs_error_codes[.]h' \
 		| grep -v '^external/' \
+		| grep -v 'src/nfs-ganesha' \
 		| while read file; do verify_file "$file"; done
