@@ -17,12 +17,12 @@ expect_equals "1 150.000 30 blkio
 unclassified 301" "$(status)"
 
 printf "subsystem blkio\nlimit /aaa 203\nlimit /aaa/bbb 504\nlimit unclassified 301\n" > "$iolimits"
-lizardfs_master_daemon reload
+lizardfs_admin_master reload-config
 expect_equals "2 150.000 30 blkio
 /aaa 203
 /aaa/bbb 504
 unclassified 301" "$(status)"
 
 printf "" > "$iolimits"
-lizardfs_master_daemon reload
+lizardfs_admin_master reload-config
 expect_equals "" "$(status)"
