@@ -4,6 +4,7 @@
 # is created and it contains information about the filestystem
 setup_local_empty_lizardfs() {
 	local use_moosefs=${USE_MOOSEFS:-}
+	local use_lizardfsXX=${LIZARDFSXX_TAG:-}
 	local use_ramdisk=${USE_RAMDISK:-}
 	local use_loop=${USE_LOOP_DISKS:-}
 	local number_of_masterservers=${MASTERSERVERS:-1}
@@ -11,24 +12,42 @@ setup_local_empty_lizardfs() {
 	local number_of_mounts=${MOUNTS:-1}
 	local disks_per_chunkserver=${DISK_PER_CHUNKSERVER:-1}
 	local auto_shadow_master=${AUTO_SHADOW_MASTER:-YES}
+	local cgi_server=${CGI_SERVER:-NO}
 	local ip_address=$(get_ip_addr)
 	local etcdir=$TEMP_DIR/mfs/etc
 	local vardir=$TEMP_DIR/mfs/var
 	local mntdir=$TEMP_DIR/mnt
+	local master_start_param=${MASTER_START_PARAM:-}
+	local shadow_start_param=${SHADOW_START_PARAM:-}
 	declare -gA lizardfs_info_
 	lizardfs_info_[chunkserver_count]=$number_of_chunkservers
+	lizardfs_info_[admin_password]=${ADMIN_PASSWORD:-password}
+
+	# Try to enable core dumps if possible
+	if [[ $(ulimit -c) == 0 ]]; then
+		ulimit -c unlimited || ulimit -c 100000000 || ulimit -c 1000000 || ulimit -c 10000 || :
+	fi
 
 	# Prepare directories for LizardFS
 	mkdir -p "$etcdir" "$vardir"
 
+	use_new_goal_config="true"
 	local oldpath="$PATH"
 	if [[ $use_moosefs ]]; then
+		use_new_goal_config="false"
 		export PATH="$MOOSEFS_DIR/bin:$MOOSEFS_DIR/sbin:$PATH"
 		build_moosefs
 	fi
 
+	if [[ $use_lizardfsXX ]]; then
+		use_new_goal_config="false"
+		LIZARDFSXX_DIR=${LIZARDFSXX_DIR_BASE}/lizardfs-${LIZARDFSXX_TAG}
+		export PATH="${LIZARDFSXX_DIR}/bin:${LIZARDFSXX_DIR}/sbin:$PATH"
+		build_lizardfsXX
+	fi
+
 	# Prepare configuration for metadata servers
-	prepare_common_metadata_server_files_
+	use_new_format=$use_new_goal_config prepare_common_metadata_server_files_
 	add_metadata_server_ 0 "master"
 	for ((msid=1 ; msid<number_of_masterservers; ++msid)); do
 		add_metadata_server_ $msid "shadow"
@@ -39,7 +58,7 @@ setup_local_empty_lizardfs() {
 	lizardfs_info_[masterserver_count]=$number_of_masterservers
 
 	# Start one masterserver with personality master
-	lizardfs_master_daemon start
+	lizardfs_master_daemon start ${master_start_param}
 
 	# Prepare the metalogger, so that any test can start it
 	prepare_metalogger_
@@ -72,13 +91,17 @@ setup_local_empty_lizardfs() {
 	# Add shadow master if not present (and not disabled); wait for it to synchronize
 	if [[ $auto_shadow_master == YES && $number_of_masterservers == 1 ]]; then
 		add_metadata_server_ auto "shadow"
-		lizardfs_master_n auto start
+		lizardfs_master_n auto start ${shadow_start_param}
 		assert_eventually 'lizardfs_shadow_synchronized auto'
 	fi
 
+	if [[ $cgi_server == YES ]]; then
+		add_cgi_server_
+	fi
+
 	# Wait for chunkservers (use lizardfs-probe only for LizardFS -- MooseFS doesn't support it)
 	if [[ ! $use_moosefs ]]; then
-		lizardfs_wait_for_ready_chunkservers $number_of_chunkservers
+		lizardfs_wait_for_all_ready_chunkservers
 	else
 		sleep 3 # A reasonable fallback
 	fi
@@ -94,24 +117,28 @@ setup_local_empty_lizardfs() {
 
 # lizardfs_chunkserver_daemon <id> start|stop|restart|kill|tests|isalive|...
 lizardfs_chunkserver_daemon() {
-	mfschunkserver -c "${lizardfs_info_[chunkserver${1}_config]}" "$2" | cat
+	local id=$1
+	shift
+	mfschunkserver -c "${lizardfs_info_[chunkserver${id}_cfg]}" "$@" | cat
 	return ${PIPESTATUS[0]}
 }
 
 lizardfs_master_daemon() {
-	mfsmaster -c "${lizardfs_info_[master${lizardfs_info_[current_master]}_cfg]}" "$1" | cat
+	mfsmaster -c "${lizardfs_info_[master${lizardfs_info_[current_master]}_cfg]}" "$@" | cat
 	return ${PIPESTATUS[0]}
 }
 
 # lizardfs_master_daemon start|stop|restart|kill|tests|isalive|...
 lizardfs_master_n() {
-	mfsmaster -c "${lizardfs_info_[master${1}_cfg]}" "$2" | cat
+	local id=$1
+	shift
+	mfsmaster -c "${lizardfs_info_[master${id}_cfg]}" "$@" | cat
 	return ${PIPESTATUS[0]}
 }
 
 # lizardfs_metalogger_daemon start|stop|restart|kill|tests|isalive|...
 lizardfs_metalogger_daemon() {
-	mfsmetalogger -c "${lizardfs_info_[metalogger_cfg]}" "$1" | cat
+	mfsmetalogger -c "${lizardfs_info_[metalogger_cfg]}" "$@" | cat
 	return ${PIPESTATUS[0]}
 }
 
@@ -153,6 +180,34 @@ create_mfsexports_cfg_() {
 	done
 }
 
+create_mfsgoals_cfg_() {
+	local goal_name
+
+	for i in {1..5}; do
+		echo "${MASTER_CUSTOM_GOALS:-}" | tr '|' '\n' | grep "^$i " | cat
+	done
+	for i in {6..20}; do
+		wildcards=
+		for ((j=0; j<i; j++)); do
+			wildcards="$wildcards _"
+		done;
+		(echo "${MASTER_CUSTOM_GOALS:-}" | tr '|' '\n' | grep "^$i ") || echo "$i $i: $wildcards"
+	done;
+	if [[ $use_new_format == "true" ]]; then
+		for i in {2..9}; do
+			echo "$((21 + $i)) xor$i: \$xor$i"
+		done;
+		for i in {2..4}; do
+			echo "$((31 + 2*(i - 2) + 0)) ec$i$((i - 1)): \$ec($i,$((i - 1)))"
+			echo "$((31 + 2*(i - 2) + 1)) ec$i$i: \$ec($i,$i)"
+		done
+	fi;
+}
+
+create_mfstopology_cfg_() {
+	echo '# empty topology...'
+}
+
 # Creates MAGIC_DEBUG_LOG which will cause test to fail is some error is logged by any daemon
 create_magic_debug_log_entry_() {
 	local servername=$1
@@ -160,17 +215,7 @@ create_magic_debug_log_entry_() {
 	# By default, fail on all prefixes passed in DEBUG_LOG_FAIL_ON
 	local prefixes=${DEBUG_LOG_FAIL_ON:-}
 
-	# This is a list of other entries, which are added to each test (but can be disabled)
-	# Add all these entries to the 'prefixes' list if not disabled using DEBUG_LOG_DISABLE_FAIL_ON
-	local auto_prefixes=("fatal.assert" "fatal.abort" "master.mismatch")
-	local disable_regex=${DEBUG_LOG_DISABLE_FAIL_ON:-$^} # default value matches nothing
 	local prefix
-	for prefix in "${auto_prefixes[@]}"; do
-		if ! [[ $prefix =~ $disable_regex ]]; then
-			prefixes+=" $prefix"
-		fi
-	done
-
 	# Create MAGIC_DEBUG_LOG_C config entry from all requested prefixes
 	if [[ $prefixes ]]; then
 		echo -n "MAGIC_DEBUG_LOG_C = "
@@ -181,6 +226,15 @@ create_magic_debug_log_entry_() {
 	fi | sed -e 's/,$//'
 }
 
+# Sometimes use Berkley DB name storage
+create_bdb_name_storage_entry_() {
+	if (($RANDOM % 2)); then
+		echo "USE_BDB_FOR_NAME_STORAGE = 0"
+	else
+		echo "USE_BDB_FOR_NAME_STORAGE = 1"
+	fi
+}
+
 create_mfsmaster_master_cfg_() {
 	local this_module_cfg_variable="MASTER_${masterserver_id}_EXTRA_CONFIG"
 	echo "PERSONALITY = master"
@@ -188,17 +242,19 @@ create_mfsmaster_master_cfg_() {
 	echo "WORKING_USER = $(id -nu)"
 	echo "WORKING_GROUP = $(id -ng)"
 	echo "EXPORTS_FILENAME = ${lizardfs_info_[master_exports]}"
-	if [[ ${lizardfs_info_[master_custom_goals]:-} ]]; then
-		echo "CUSTOM_GOALS_FILENAME = ${lizardfs_info_[master_custom_goals]}"
-	fi
+	echo "TOPOLOGY_FILENAME = ${lizardfs_info_[master_topology]}"
+	echo "CUSTOM_GOALS_FILENAME = ${lizardfs_info_[master_custom_goals]}"
 	echo "DATA_PATH = $masterserver_data_path"
 	echo "MATOML_LISTEN_PORT = ${lizardfs_info_[matoml]}"
 	echo "MATOCS_LISTEN_PORT = ${lizardfs_info_[matocs]}"
 	echo "MATOCL_LISTEN_PORT = ${lizardfs_info_[matocl]}"
+	echo "MATOTS_LISTEN_PORT = ${lizardfs_info_[matots]}"
 	echo "METADATA_CHECKSUM_INTERVAL = 1"
+	echo "ADMIN_PASSWORD = ${lizardfs_info_[admin_password]}"
 	create_magic_debug_log_entry_ "master_${masterserver_id}"
 	echo "${MASTER_EXTRA_CONFIG-}" | tr '|' '\n'
 	echo "${!this_module_cfg_variable-}" | tr '|' '\n'
+	create_bdb_name_storage_entry_
 }
 
 create_mfsmaster_shadow_cfg_() {
@@ -208,19 +264,21 @@ create_mfsmaster_shadow_cfg_() {
 	echo "WORKING_USER = $(id -nu)"
 	echo "WORKING_GROUP = $(id -ng)"
 	echo "EXPORTS_FILENAME = ${lizardfs_info_[master_exports]}"
-	if [[ ${lizardfs_info_[master_custom_goals]:-} ]]; then
-		echo "CUSTOM_GOALS_FILENAME = ${lizardfs_info_[master_custom_goals]}"
-	fi
+	echo "TOPOLOGY_FILENAME = ${lizardfs_info_[master_topology]}"
+	echo "CUSTOM_GOALS_FILENAME = ${lizardfs_info_[master_custom_goals]}"
 	echo "DATA_PATH = $masterserver_data_path"
 	echo "MATOML_LISTEN_PORT = $masterserver_matoml_port"
 	echo "MATOCS_LISTEN_PORT = $masterserver_matocs_port"
 	echo "MATOCL_LISTEN_PORT = $masterserver_matocl_port"
+	echo "MATOTS_LISTEN_PORT = $masterserver_matots_port"
 	echo "MASTER_HOST = $(get_ip_addr)"
 	echo "MASTER_PORT = ${lizardfs_info_[matoml]}"
 	echo "METADATA_CHECKSUM_INTERVAL = 1"
+	echo "ADMIN_PASSWORD = ${lizardfs_info_[admin_password]}"
 	create_magic_debug_log_entry_ "shadow_${masterserver_id}"
 	echo "${MASTER_EXTRA_CONFIG-}" | tr '|' '\n'
 	echo "${!this_module_cfg_variable-}" | tr '|' '\n'
+	create_bdb_name_storage_entry_
 }
 
 lizardfs_make_conf_for_shadow() {
@@ -231,7 +289,7 @@ lizardfs_make_conf_for_shadow() {
 lizardfs_make_conf_for_master() {
 	local new_master=$1
 	local old_master=${lizardfs_info_[current_master]}
-	# move master responsiblity to new masterserver
+	# move master responsibility to new masterserver
 	cp -f "${lizardfs_info_[master${new_master}_master_cfg]}" "${lizardfs_info_[master${new_master}_cfg]}"
 	lizardfs_info_[master_cfg]=${lizardfs_info_[master${new_master}_master_cfg]}
 	lizardfs_info_[master_data_path]=${lizardfs_info_[master${new_master}_data_path]}
@@ -244,14 +302,15 @@ lizardfs_current_master_id() {
 
 prepare_common_metadata_server_files_() {
 	create_mfsexports_cfg_ > "$etcdir/mfsexports.cfg"
+	create_mfstopology_cfg_ > "$etcdir/mfstopology.cfg"
+	create_mfsgoals_cfg_ > "$etcdir/mfsgoals.cfg"
 	lizardfs_info_[master_exports]="$etcdir/mfsexports.cfg"
-	if [[ ${MASTER_CUSTOM_GOALS:-} ]]; then
-		echo "$MASTER_CUSTOM_GOALS" | tr '|' '\n' > "$etcdir/goals.cfg"
-		lizardfs_info_[master_custom_goals]="$etcdir/goals.cfg"
-	fi
+	lizardfs_info_[master_topology]="$etcdir/mfstopology.cfg"
+	lizardfs_info_[master_custom_goals]="$etcdir/mfsgoals.cfg"
 	get_next_port_number "lizardfs_info_[matoml]"
 	get_next_port_number "lizardfs_info_[matocl]"
 	get_next_port_number "lizardfs_info_[matocs]"
+	get_next_port_number "lizardfs_info_[matots]"
 }
 
 add_metadata_server_() {
@@ -261,6 +320,7 @@ add_metadata_server_() {
 	local masterserver_matoml_port
 	local masterserver_matocl_port
 	local masterserver_matocs_port
+	local masterserver_matots_port
 	local masterserver_data_path=$vardir/master${masterserver_id}
 	local masterserver_master_cfg=$etcdir/mfsmaster${masterserver_id}_master.cfg
 	local masterserver_shadow_cfg=$etcdir/mfsmaster${masterserver_id}_shadow.cfg
@@ -269,6 +329,7 @@ add_metadata_server_() {
 	get_next_port_number masterserver_matoml_port
 	get_next_port_number masterserver_matocl_port
 	get_next_port_number masterserver_matocs_port
+	get_next_port_number masterserver_matots_port
 	mkdir "$masterserver_data_path"
 	create_mfsmaster_master_cfg_ > "$masterserver_master_cfg"
 	create_mfsmaster_shadow_cfg_ > "$masterserver_shadow_cfg"
@@ -289,6 +350,7 @@ add_metadata_server_() {
 	lizardfs_info_[master${masterserver_id}_matoml]=$masterserver_matoml_port
 	lizardfs_info_[master${masterserver_id}_matocl]=$masterserver_matocl_port
 	lizardfs_info_[master${masterserver_id}_matocs]=$masterserver_matocs_port
+	lizardfs_info_[master${masterserver_id}_matots]=$masterserver_matocs_port
 }
 
 create_mfsmetalogger_cfg_() {
@@ -310,8 +372,14 @@ create_mfshdd_cfg_() {
 	local n=$disks_per_chunkserver
 	if [[ $use_ramdisk ]]; then
 		local disk_number
-		for disk_number in $(seq 1 $n); do
-			local disk_dir=$RAMDISK_DIR/hdd_${chunkserver_id}_${disk_number}
+		for (( disk_number=0; disk_number<n; disk_number++ )); do
+			# Use path provided in env variable, if present generate some pathname otherwise.
+			local this_disk_variable="CHUNKSERVER_${chunkserver_id}_DISK_${disk_number}"
+			if [[ ${!this_disk_variable-} ]]; then
+				local disk_dir=${!this_disk_variable}
+			else
+				local disk_dir=$RAMDISK_DIR/hdd_${chunkserver_id}_${disk_number}
+			fi
 			mkdir -pm 777 $disk_dir
 			echo $disk_dir
 		done
@@ -337,6 +405,7 @@ create_mfschunkserver_cfg_() {
 	echo "WORKING_GROUP = $(id -ng)"
 	echo "DATA_PATH = $chunkserver_data_path"
 	echo "HDD_CONF_FILENAME = $hdd_cfg"
+	echo "HDD_LEAVE_SPACE_DEFAULT = 128MiB"
 	echo "MASTER_HOST = $ip_address"
 	echo "MASTER_PORT = ${lizardfs_info_[matocs]}"
 	echo "CSSERV_LISTEN_PORT = $csserv_port"
@@ -346,6 +415,16 @@ create_mfschunkserver_cfg_() {
 	echo "${!this_module_cfg_variable-}" | tr '|' '\n'
 }
 
+# Run every second chunkserver with each chunk format
+chunkserver_chunk_format_cfg_() {
+	local chunkserver_id=$1
+	if (($RANDOM % 2)); then
+		echo "CREATE_NEW_CHUNKS_IN_MOOSEFS_FORMAT = 0"
+	else
+		echo "CREATE_NEW_CHUNKS_IN_MOOSEFS_FORMAT = 1"
+	fi
+}
+
 add_chunkserver_() {
 	local chunkserver_id=$1
 	local csserv_port
@@ -356,11 +435,14 @@ add_chunkserver_() {
 	get_next_port_number csserv_port
 	create_mfshdd_cfg_ > "$hdd_cfg"
 	create_mfschunkserver_cfg_ > "$chunkserver_cfg"
+	# If the chunk format wasn't yet chosen set it:
+	grep CREATE_NEW_CHUNKS_IN_MOOSEFS_FORMAT "$chunkserver_cfg" || \
+		chunkserver_chunk_format_cfg_ $chunkserver_id >> "$chunkserver_cfg"
 	mkdir -p "$chunkserver_data_path"
 	mfschunkserver -c "$chunkserver_cfg" start
 
 	lizardfs_info_[chunkserver${chunkserver_id}_port]=$csserv_port
-	lizardfs_info_[chunkserver${chunkserver_id}_config]=$chunkserver_cfg
+	lizardfs_info_[chunkserver${chunkserver_id}_cfg]=$chunkserver_cfg
 	lizardfs_info_[chunkserver${chunkserver_id}_hdd]=$hdd_cfg
 }
 
@@ -379,12 +461,17 @@ create_mfsmount_cfg_() {
 
 do_mount_() {
 	local mount_id=$1
+	local workingdir=$TEMP_DIR/var/mount$mount_id
+	# Make sure mount process is in a writable directory (for core dumps)
+	mkdir -p $workingdir
+	cd $workingdir
 	for try in $(seq 1 $max_tries); do
 		${lizardfs_info_[mntcall${mount_id}]} && return 0
 		echo "Retrying in 1 second ($try/$max_tries)..."
 		sleep 1
 	done
 	echo "Cannot mount in $mount_dir, exiting"
+	cd -
 	exit 2
 }
 
@@ -395,7 +482,7 @@ add_mount_() {
 	create_mfsmount_cfg_ ${mount_id} > "$mount_cfg"
 	mkdir -p "$mount_dir"
 	lizardfs_info_[mount${mount_id}]="$mount_dir"
-	lizardfs_info_[mount${mount_id}_config]="$mount_cfg"
+	lizardfs_info_[mount${mount_id}_cfg]="$mount_cfg"
 	max_tries=30
 	fuse_options=""
 	for fuse_option in $(echo ${FUSE_EXTRA_CONFIG-} | tr '|' '\n'); do
@@ -404,21 +491,31 @@ add_mount_() {
 				|| test_fail "Your libfuse doesn't support $fuse_option_name flag"
 		fuse_options+="-o $fuse_option "
 	done
-	local call="${command_prefix} mfsmount -c ${mount_cfg} ${mount_dir} ${fuse_options}"
+	local call="${command_prefix} mfsmount -o big_writes -c ${mount_cfg} ${mount_dir} ${fuse_options}"
 	lizardfs_info_[mntcall$mount_id]=$call
 	do_mount_ ${mount_id}
 }
 
+add_cgi_server_() {
+	local cgi_server_port
+	local pidfile="$vardir/lizardfs-cgiserver.pid"
+	get_next_port_number cgi_server_port
+	lizardfs-cgiserver -P "$cgi_server_port" -p "$pidfile"
+	lizardfs_info_[cgi_pidfile]=$pidfile
+	lizardfs_info_[cgi_port]=$cgi_server_port
+	lizardfs_info_[cgi_url]="http://localhost:$cgi_server_port/mfs.cgi?masterport=${lizardfs_info_[matocl]}"
+}
+
 # Some helper functions for tests to manipulate the existing installation
 
 mfs_dir_info() {
 	if (( $# != 2 )); then
-		echo "Incorrect usage of mfs_dir_info with args $*";
+		echo "Incorrect usage of lizardfs dir_info with args $*";
 		exit 2;
 	fi;
 	field=$1
 	file=$2
-	mfsdirinfo "$file" | grep -w "$field" | grep -o '[0-9]*'
+	lizardfs dirinfo "$file" | grep -w "$field" | grep -o '[0-9]*'
 }
 
 find_first_chunkserver_with_chunks_matching() {
@@ -427,7 +524,7 @@ find_first_chunkserver_with_chunks_matching() {
 	local chunkserver
 	for (( chunkserver=0 ; chunkserver < count ; ++chunkserver )); do
 		local hdds=$(cat "${lizardfs_info_[chunkserver${chunkserver}_hdd]}")
-		if [[ $(find $hdds -name "$pattern") ]]; then
+		if [[ $(find $hdds -type f -name "$pattern") ]]; then
 			echo $chunkserver
 			return 0
 		fi
@@ -439,12 +536,12 @@ find_first_chunkserver_with_chunks_matching() {
 find_chunkserver_chunks() {
 	local chunkserver_number=$1
 	shift
-	local hdds=$(sed -e 's|$|/[A-F0-9][A-F0-9]/|' \
+	local hdds=$(sed -e 's/*//' -e 's|$|/chunks[A-F0-9][A-F0-9]/|' \
 			"${lizardfs_info_[chunkserver${chunkserver_number}_hdd]}")
 	if (( $# > 0 )); then
-		find $hdds -name "chunk*.mfs" -a "(" "$@" ")"
+		find $hdds "(" -name 'chunk*.liz' -o -name 'chunk*.mfs' ")" -a "(" "$@" ")"
 	else
-		find $hdds -name "chunk*.mfs"
+		find $hdds "(" -name 'chunk*.liz' -o -name 'chunk*.mfs' ")"
 	fi
 }
 
@@ -453,12 +550,7 @@ find_all_chunks() {
 	local count=${lizardfs_info_[chunkserver_count]}
 	local chunkserver
 	for (( chunkserver=0 ; chunkserver < count ; ++chunkserver )); do
-		local hdds=$(sed -e 's|$|/[A-F0-9][A-F0-9]/|' "${lizardfs_info_[chunkserver${chunkserver}_hdd]}")
-		if (( $# > 0 )); then
-			find $hdds -name "chunk*.mfs" -a "(" "$@" ")"
-		else
-			find $hdds -name "chunk*.mfs"
-		fi
+		find_chunkserver_chunks $chunkserver "$@"
 	done
 }
 
@@ -472,6 +564,40 @@ lizardfs_probe_master() {
 	lizardfs-probe "$command" localhost "${lizardfs_info_[matocl]}" --porcelain "$@"
 }
 
+# A useful shortcut for lizardfs-admin commands which require authentication
+# Usage: lizardfs_admin_master <command> [option...]
+# Calls lizardfs-admin with the given command and and automatically adds address
+# of the master server and authenticates
+lizardfs_admin_master() {
+	local command="$1"
+	shift
+	local port=${lizardfs_info_[matocl]}
+	lizardfs-admin "$command" localhost "$port" "$@" <<< "${lizardfs_info_[admin_password]}"
+}
+
+# A useful shortcut for lizardfs-admin commands which require authentication
+# Usage: lizardfs_admin_shadow <n> <command> [option...]
+# Calls lizardfs-admin with the given command and and automatically adds address
+# of the n'th shadow master server and authenticates
+lizardfs_admin_shadow() {
+	local id="$1"
+	local command="$2"
+	shift 2
+	local port=${lizardfs_info_[master${id}_matocl]}
+	lizardfs-admin "$command" localhost "$port" "$@" <<< "${lizardfs_info_[admin_password]}"
+}
+
+# Stops the active master server without dumping metadata
+lizardfs_stop_master_without_saving_metadata() {
+	lizardfs_admin_master stop-master-without-saving-metadata
+	assert_eventually "! mfsmaster -c ${lizardfs_info_[master_cfg]} isalive"
+}
+
+# print the number of fully operational chunkservers
+lizardfs_ready_chunkservers_count() {
+	lizardfs-probe ready-chunkservers-count localhost ${lizardfs_info_[matocl]}
+}
+
 # lizardfs_wait_for_ready_chunkservers <num> -- waits until <num> chunkservers are fully operational
 lizardfs_wait_for_ready_chunkservers() {
 	local chunkservers=$1
@@ -501,12 +627,9 @@ lizardfs_shadow_synchronized() {
 }
 
 # Prints number of chunks on each chunkserver in the following form:
-# <ip1>:<port1> <chunks1>
-# <ip2>:<port2> <chunks2>
+# <ip1>:<port1>:<label> <chunks1>
+# <ip2>:<port2>:<label> <chunks2>
 # ...
 lizardfs_rebalancing_status() {
-	lizardfs_probe_master list-chunkservers | sort | awk '$2 == "'$LIZARDFS_VERSION'" {print $1,$3}'
+	lizardfs_probe_master list-chunkservers | sort | awk '$2 == "'$LIZARDFS_VERSION'" {print $1":"$10,$3}'
 }
-
-LIZARDFS_BLOCK_SIZE=$((64 * 1024))
-LIZARDFS_CHUNK_SIZE=$((1024 * LIZARDFS_BLOCK_SIZE))
